/***************************************************************************
                          controller.cpp  -  description
                             -------------------
    begin                : Thu Apr 29 2004
    copyright            : (C) 2004 by Michael Rosskopf
    email                : rosskopf@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "controller.h"

// Constructors and destructors
Controller::Controller(char* outputFile, char* configFile, double rndModifier){

  // Initialization

  // Set Parameters (The default values are setted by the parameter constructor)
  programParameters = new ProgramParameters(outputFile,configFile);

  // Create a random number generator
  rnd = new RandomNumberGenerator();
  rnd->initWithModifier(rndModifier); // <--- Initialization of the random number generator is important

  // Create a random number generator for gamma distributions
  gammarnd = new Random();
  gammarnd->start_kiss(-(time(0)) + rndModifier);

  // Create a threetree
  threeTree = new ThreeTree();

  // Load a default model parameters parameters
  modelParameters = new ModelParameters();
  modelParameters->loadModelParameters(programParameters->getModelParameterFilename());

  // Create an object for analytic solutions
  analyticEstimator = new AnalyticEstimator();

  // Create an optimizer
  optimizer = new Optimizer(programParameters);

  // Create an optimizer
  mlOptimizer = new MLOptimizer();

  // Create a best individual collector
  resultCollector = new ResultCollector();
  resultCollector->init(programParameters->getNumberOfRuns());

  // Precision of the output
  cout.precision(programParameters->getOutputPrecision());

  // Check program parameters
  bool valid = true;
  valid = programParameters->checkProgramParameters();

  // Initialize data and load if a file is used
  data = new Data();
  if(valid && (programParameters->getMode() == 2 || programParameters->getMode() == 3)){
    data->load(programParameters->getDataTableFilename());
    // Check whether the data set contains the samples
    valid = data->dataCheck(programParameters);
  }

  // Call of the selected tool if program parameters are valid
  if(valid){
    if(programParameters->getMode() == 1) performDataGeneration(outputFile);
    if(programParameters->getMode() == 2) performEstimationOnRealData(outputFile);
    if(programParameters->getMode() == 3) performBootstrappingSeries(outputFile);
    if(programParameters->getMode() == 4) performGenerateEstimateSeries(outputFile);
  }    
}
Controller::~Controller(){
  delete data;
  delete optimizer;
  delete analyticEstimator;
  delete modelParameters;
  delete threeTree;
  delete gammarnd;
  delete rnd;
  delete programParameters;
}

// performDataGeneration: Implements data generation
void Controller::performDataGeneration(char* outputFile){

  int sizeSample1 = 0;
  int sizeSample2 = 0;
  int sizeSample3 = 0;
  int i;

  // Set sample labels to 1, 2, 3 for data generation
  programParameters->setSample1(1);
  programParameters->setSample2(2);
  if(programParameters->getSample3() != 0) programParameters->setSample3(3);

  cout << "\nData generation" << endl;
  cout << "Insert number of individuals in sample 1: ";
  cin >> sizeSample1;
  cout << "Insert number of individuals in sample 2: ";
  cin >> sizeSample2;
  if(programParameters->getSample3() != 0){
    cout << "Insert number of individuals in sample 3: ";
    cin >> sizeSample3;
  } else {
    sizeSample3 = 1;
  }

  Data* artificialData = new Data();
  artificialData->init(sizeSample1 + sizeSample2 + sizeSample3,programParameters->getNumberOfGenes(),3);
  artificialData->setSampleStart(0,0);
  artificialData->setSampleSize(0,sizeSample1);
  artificialData->setSampleStart(1,sizeSample1);
  artificialData->setSampleSize(1,sizeSample2);
  artificialData->setSampleStart(2,sizeSample1 + sizeSample2);
  artificialData->setSampleSize(2,sizeSample3);
  for(i = 0; i < sizeSample1; i++){
    artificialData->setLabel(i,1);
  }
  for(i = 0; i < sizeSample2; i++){
    artificialData->setLabel(artificialData->getSampleStart(1) + i,2);
  }
  for(i = 0; i < sizeSample3; i++){
    artificialData->setLabel(artificialData->getSampleStart(2) + i,3);
  }
  threeTree->generation(programParameters,modelParameters,artificialData,rnd,gammarnd);
  artificialData->toString();
  artificialData->save(outputFile);

  DistributionMeasures* distributionMeasures = new DistributionMeasures();
  artificialData->calcDistributionMeasuresFromMeanOfAllPairs(programParameters,distributionMeasures,true);
  cout << "\nMoments of generated data:" << endl;
  distributionMeasures->toString(programParameters);
}

// performEstimationOnRealData: Implements parameter estimation on real data
void Controller::performEstimationOnRealData(char* outputFile){

  // Initializing
  GraphicTools gs = GraphicTools();

  cout << "\nNumber of genes in table: " << data->getNumberOfGenes() << endl;
  DistributionMeasures* targetValues = new DistributionMeasures();
  double* targetDistribution_1_2 = new double[data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample2() - 1) * data->getNumberOfGenes()];
  double* targetHistogramPlot_1_2 = new double[programParameters->getNumberOfBins()];
  double* targetDistribution_1_3 = new double[data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes()];
  double* targetHistogramPlot_1_3 = new double[programParameters->getNumberOfBins()];
  double* targetDistribution_2_3 = new double[data->getSampleSize(programParameters->getSample2() - 1) * data->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes()];
  double* targetHistogramPlot_2_3 = new double[programParameters->getNumberOfBins()];
  double* targetDistribution2Samples = new double[data->getNumberOfGenes()];
  double* targetHistogramPlot2Samples = new double[programParameters->getNumberOfBins()];

  double lowerBound_1_2 = 99999999;
  double upperBound_1_2 = -99999999;
  double lowerBound_1_3 = 99999999;
  double upperBound_1_3 = -99999999;
  double lowerBound_2_3 = 99999999;
  double upperBound_2_3 = -99999999;
  double lowerBound2Samples = 99999999;
  double upperBound2Samples = -99999999;
  long indexBuffer;

  if(programParameters->getAnalysisType() != 2){
    // AnalysisType == 1 (Mean of all pairwise comparisons)

    // Calculate target distributions
    if(programParameters->getNonMutationalEffects() == 2){
      cout << "Calculate target distributions" << endl;  
      data->calcTargetDistribution(targetDistribution_1_2,programParameters->getSample1() - 1,programParameters->getSample2() - 1);
      if(programParameters->getSample3() != 0) data->calcTargetDistribution(targetDistribution_1_3,programParameters->getSample1() - 1,programParameters->getSample3() - 1);
      if(programParameters->getSample3() != 0) data->calcTargetDistribution(targetDistribution_2_3,programParameters->getSample2() - 1,programParameters->getSample3() - 1);
    }
    cout << "\nMoments estimated from the dataset:" << endl;
    data->calcDistributionMeasuresFromMeanOfAllPairs(programParameters,targetValues,true);
    targetValues->toString(programParameters);
    gs.calcHistogramPlot(targetDistribution_1_2,data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample2() - 1) * data->getNumberOfGenes(),targetHistogramPlot_1_2,programParameters->getNumberOfBins(),&lowerBound_1_2,&upperBound_1_2);
    if(programParameters->getSample3() != 0) gs.calcHistogramPlot(targetDistribution_1_3,data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes(),targetHistogramPlot_1_3,programParameters->getNumberOfBins(),&lowerBound_1_3,&upperBound_1_3);
    if(programParameters->getSample3() != 0) gs.calcHistogramPlot(targetDistribution_2_3,data->getSampleSize(programParameters->getSample2() - 1) * data->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes(),targetHistogramPlot_2_3,programParameters->getNumberOfBins(),&lowerBound_2_3,&upperBound_2_3);

    for(long runNumber = 0; runNumber < programParameters->getNumberOfRuns(); runNumber++){

      if(programParameters->getMED() == 3 || programParameters->getNonMutationalEffects() == 2){
        // Model which needs optimisation for parameter estimation
        cout << "\nOptimizer run " << runNumber + 1 << "/" << programParameters->getNumberOfRuns() << endl;
        if(programParameters->getNonMutationalEffects() == 1){
          // M-gamma model
          optimizer->estimateMGammaParameters(programParameters,data,threeTree,targetValues,rnd,gammarnd);
        }
        if(programParameters->getNonMutationalEffects() == 2){
          if(programParameters->getMED() == 1){
            // M&E-normal model
            if(programParameters->getMethod() == 1) optimizer->estimateMENormalParameters(programParameters,data,threeTree,targetHistogramPlot_1_2,lowerBound_1_2,upperBound_1_2,targetValues,rnd,gammarnd); // Chi^2
            if(programParameters->getMethod() == 2) mlOptimizer->estimateModelParametersMENormal(programParameters,data,rnd,gammarnd); // ML
          }
          if(programParameters->getMED() == 2){
            // M&E-extreme model
            optimizer->estimateMEExtremeParameters(programParameters,data,threeTree,targetHistogramPlot_1_2,lowerBound_1_2,upperBound_1_2,targetValues,rnd,gammarnd);
          }
        }
        cout << "\n----------------------------------------------------------------" << endl;
        cout << "\nResult:" << endl;
        // Get the results and store them in the result collector
        if(programParameters->getMethod() == 1 || programParameters->getMED() == 3){
          optimizer->getEstimate()->toString(programParameters);
          if(optimizer->getEstimate()->getQualityValue() != 1.0){
            cout << "Limit reached. No solution!" << endl;
          }
          resultCollector->setModelParameters(runNumber,*optimizer->getEstimate(),*targetValues);
          resultCollector->saveModelParameters(runNumber,outputFile);
        }
        if(programParameters->getMethod() == 2 && programParameters->getMED() != 3){
          mlOptimizer->getEstimate()->toString(programParameters);
          if(mlOptimizer->getEstimate()->getQualityValue() != 1.0){
            cout << "Limit reached. No solution!" << endl;
          }
          resultCollector->setModelParameters(runNumber,*mlOptimizer->getEstimate(),*targetValues);
          resultCollector->saveModelParameters(runNumber,outputFile);
        }

      } else {
        // Model whose parameters can be estimated analytically
        analyticEstimator->calculateEstimate(programParameters,targetValues);
        cout << "\nResult:" << endl;
        analyticEstimator->getEstimate()->toString(programParameters);
        // Get the results and store them in the result collector
        resultCollector->setModelParameters(runNumber,*analyticEstimator->getEstimate(),*targetValues);
      }
      if(programParameters->getOutputMode() > 1){
        cout << "\n----------------------------------------------------------------" << endl;
      }
    }

    // Save results
    cout << "\nSave table in file " << outputFile << endl;
    resultCollector->saveCompleteParameterTable(outputFile,false);

  } else {
    // AnalysisType == 2 (Estimation for all pairs of individuals)

    ModelParameters* meanWithinGroup1 = new ModelParameters();
    ModelParameters* meanWithinGroup2 = new ModelParameters();
    ModelParameters* meanBetweenGroups = new ModelParameters();
    double meanWithinGroup1Var = 0.0;
    double meanWithinGroup1Skew = 0.0;
    double meanWithinGroup1Kurt = 0.0;
    double meanWithinGroup2Var = 0.0;
    double meanWithinGroup2Skew = 0.0;
    double meanWithinGroup2Kurt = 0.0;
    double meanBetweenGroupsVar = 0.0;
    double meanBetweenGroupsSkew = 0.0;
    double meanBetweenGroupsKurt = 0.0;

    ofstream file(outputFile);
    file.precision(programParameters->getOutputPrecision());
    if(programParameters->getMED() == 1) file << "Pair\tsigma_m\tsigma_e\td\t***\tVar12\tSkew12\tKurt12" << endl;
    if(programParameters->getMED() == 2) file << "Pair\tbeta\tsigma_e\td_1\td_2\t***\tVar12\tSkew12\tKurt12" << endl;
    cout << "\n----------------------------------------------------------------" << endl;
    cout << "Estimate model parameters within group " << programParameters->getSample1() << endl;
    cout << "----------------------------------------------------------------" << endl;
    file << "\nEstimate model parameters within group " << programParameters->getSample1() << endl;
    if(data->getSampleSize(programParameters->getSample1() - 1) == 1){
      cout << "Only one sample in group " << programParameters->getSample1() << ". No comparisons possible!" << endl;
    } else {
      for(long i = 0; i < data->getSampleSize(programParameters->getSample1() - 1) - 1; i++){
        for(long j = i + 1; j < data->getSampleSize(programParameters->getSample1() - 1); j++){
          cout << "\nComparisons between sample " << i + 1 << " and sample " << j + 1 << ":" << endl;
          estimateFromTwoSamples(data,targetValues,targetDistribution2Samples,targetHistogramPlot2Samples,data->getSampleStart(programParameters->getSample1() - 1) + i,data->getSampleStart(programParameters->getSample1() - 1) + j);
          file << data->getSampleStart(programParameters->getSample1() - 1) + i + 1 << " - " << data->getSampleStart(programParameters->getSample1() - 1) + j + 1 << "\t";
          if(programParameters->getNonMutationalEffects() == 2){

            if(programParameters->getMethod() == 1){ 
              optimizer->getEstimate()->toString(programParameters);
              if(optimizer->getEstimate()->getQualityValue() == 1.0){
                if(programParameters->getMED() == 1) file << optimizer->getEstimate()->getBetaOrSigmaM() << "\t" << optimizer->getEstimate()->getSigmaE() << "\t" << optimizer->getEstimate()->getD1();
                if(programParameters->getMED() == 2) file << optimizer->getEstimate()->getBetaOrSigmaM() << "\t" << optimizer->getEstimate()->getSigmaE() << "\t" << optimizer->getEstimate()->getD1() << "\t" << optimizer->getEstimate()->getD2();
              } else {
                file << "Limit reached. No solution!";
              }
              meanWithinGroup1->addUpValues(optimizer->getEstimate());
            }
            if(programParameters->getMethod() == 2){ 
              // ML method only for the M&E-normal model
              mlOptimizer->getEstimate()->toString(programParameters);
              if(mlOptimizer->getEstimate()->getQualityValue() == 1.0){
                file << mlOptimizer->getEstimate()->getBetaOrSigmaM() << "\t" << mlOptimizer->getEstimate()->getSigmaE() << "\t" << mlOptimizer->getEstimate()->getD1();
              } else {
                file << "Limit reached. No solution!";
              }
              meanWithinGroup1->addUpValues(mlOptimizer->getEstimate());
            }

          } else {
            analyticEstimator->getEstimate()->toString(programParameters);
              if(programParameters->getMED() == 1) file << analyticEstimator->getEstimate()->getBetaOrSigmaM() << "\t" << analyticEstimator->getEstimate()->getSigmaE() << "\t" << analyticEstimator->getEstimate()->getD1();
              if(programParameters->getMED() == 2) file << analyticEstimator->getEstimate()->getBetaOrSigmaM() << "\t" << analyticEstimator->getEstimate()->getSigmaE() << "\t" << analyticEstimator->getEstimate()->getD1() << "\t" << analyticEstimator->getEstimate()->getD2();
            meanWithinGroup1->addUpValues(analyticEstimator->getEstimate());
          }
          file << "\t***\t" << targetValues->getVariance12() << "\t" << targetValues->getSkewness12() << "\t" << targetValues->getKurtosis12() << endl;
          meanWithinGroup1Var = meanWithinGroup1Var + targetValues->getVariance12();
          meanWithinGroup1Skew = meanWithinGroup1Skew + targetValues->getSkewness12();
          meanWithinGroup1Kurt = meanWithinGroup1Kurt + targetValues->getKurtosis12();
        }
      }
      meanWithinGroup1->divideValues(data->getSampleSize(programParameters->getSample1() - 1) * (data->getSampleSize(programParameters->getSample1() - 1) - 1) / 2.0);
      meanWithinGroup1Var = meanWithinGroup1Var / (data->getSampleSize(programParameters->getSample1() - 1) * (data->getSampleSize(programParameters->getSample1() - 1) - 1) / 2.0);
      meanWithinGroup1Skew = meanWithinGroup1Skew / (data->getSampleSize(programParameters->getSample1() - 1) * (data->getSampleSize(programParameters->getSample1() - 1) - 1) / 2.0);
      meanWithinGroup1Kurt = meanWithinGroup1Kurt / (data->getSampleSize(programParameters->getSample1() - 1) * (data->getSampleSize(programParameters->getSample1() - 1) - 1) / 2.0);
      file << "\nMean:\t";
      if(programParameters->getMED() == 1) file << meanWithinGroup1->getBetaOrSigmaM() << "\t" << meanWithinGroup1->getSigmaE() << "\t" << meanWithinGroup1->getD1() << "\t***\t" << meanWithinGroup1Var << "\t" << meanWithinGroup1Skew << "\t" << meanWithinGroup1Kurt << endl;
      if(programParameters->getMED() == 2) file << meanWithinGroup1->getBetaOrSigmaM() << "\t" << meanWithinGroup1->getSigmaE() << "\t" << meanWithinGroup1->getD1() << "\t" << meanWithinGroup1->getD2() << "\t***\t" << meanWithinGroup1Var << "\t" << meanWithinGroup1Skew << "\t" << meanWithinGroup1Kurt << endl;
    }

    cout << "\n----------------------------------------------------------------" << endl;
    cout << "Estimate model parameters within group " << programParameters->getSample2() << endl;
    cout << "----------------------------------------------------------------" << endl;
    file << "\nEstimate model parameters within group " << programParameters->getSample2() << endl;
    if(data->getSampleSize(programParameters->getSample2() - 1) == 1){
      cout << "Only one sample in group " << programParameters->getSample2() << ". No comparisons possible!" << endl;
    } else {
      for(long i = 0; i < data->getSampleSize(programParameters->getSample2() - 1) - 1; i++){
        for(long j = i + 1; j < data->getSampleSize(programParameters->getSample2() - 1); j++){
          cout << "\nComparisons between sample " << i + 1 << " and sample " << j + 1 << ":" << endl;
          estimateFromTwoSamples(data,targetValues,targetDistribution2Samples,targetHistogramPlot2Samples,data->getSampleStart(programParameters->getSample2() - 1) + i,data->getSampleStart(programParameters->getSample2() - 1) + j);
          file << data->getSampleStart(programParameters->getSample2() - 1) + i + 1 << " - " << data->getSampleStart(programParameters->getSample2() - 1) + j + 1 << "\t";
          if(programParameters->getNonMutationalEffects() == 2){

            if(programParameters->getMethod() == 1){ 
              optimizer->getEstimate()->toString(programParameters);
              if(optimizer->getEstimate()->getQualityValue() == 1.0){
                if(programParameters->getMED() == 1) file << optimizer->getEstimate()->getBetaOrSigmaM() << "\t" << optimizer->getEstimate()->getSigmaE() << "\t" << optimizer->getEstimate()->getD1();
                if(programParameters->getMED() == 2) file << optimizer->getEstimate()->getBetaOrSigmaM() << "\t" << optimizer->getEstimate()->getSigmaE() << "\t" << optimizer->getEstimate()->getD1() << "\t" << optimizer->getEstimate()->getD2();
              } else {
                file << "Limit reached. No solution!";
              }
              meanWithinGroup2->addUpValues(optimizer->getEstimate());
            }
            if(programParameters->getMethod() == 2){
              // ML method only for the M&E-normal model
              mlOptimizer->getEstimate()->toString(programParameters);
              if(mlOptimizer->getEstimate()->getQualityValue() == 1.0){
                file << mlOptimizer->getEstimate()->getBetaOrSigmaM() << "\t" << mlOptimizer->getEstimate()->getSigmaE() << "\t" << mlOptimizer->getEstimate()->getD1();
              } else {
                file << "Limit reached. No solution!";
              }
              meanWithinGroup2->addUpValues(mlOptimizer->getEstimate());
            }

          } else {
            analyticEstimator->getEstimate()->toString(programParameters);
            if(programParameters->getMED() == 1) file << analyticEstimator->getEstimate()->getBetaOrSigmaM() << "\t" << analyticEstimator->getEstimate()->getSigmaE() << "\t" << analyticEstimator->getEstimate()->getD1();
            if(programParameters->getMED() == 2) file << analyticEstimator->getEstimate()->getBetaOrSigmaM() << "\t" << analyticEstimator->getEstimate()->getSigmaE() << "\t" << analyticEstimator->getEstimate()->getD1() << "\t" << analyticEstimator->getEstimate()->getD2();
            meanWithinGroup2->addUpValues(analyticEstimator->getEstimate());
          }
          file << "\t***\t" << targetValues->getVariance12() << "\t" << targetValues->getSkewness12() << "\t" << targetValues->getKurtosis12() << endl;
          meanWithinGroup2Var = meanWithinGroup2Var + targetValues->getVariance12();
          meanWithinGroup2Skew = meanWithinGroup2Skew + targetValues->getSkewness12();
          meanWithinGroup2Kurt = meanWithinGroup2Kurt + targetValues->getKurtosis12();
        }
      }
      meanWithinGroup2->divideValues(data->getSampleSize(programParameters->getSample2() - 1) * (data->getSampleSize(programParameters->getSample2() - 1) - 1) / 2.0);
      meanWithinGroup2Var = meanWithinGroup2Var / (data->getSampleSize(programParameters->getSample2() - 1) * (data->getSampleSize(programParameters->getSample2() - 1) - 1) / 2.0);
      meanWithinGroup2Skew = meanWithinGroup2Skew / (data->getSampleSize(programParameters->getSample2() - 1) * (data->getSampleSize(programParameters->getSample2() - 1) - 1) / 2.0);
      meanWithinGroup2Kurt = meanWithinGroup2Kurt / (data->getSampleSize(programParameters->getSample2() - 1) * (data->getSampleSize(programParameters->getSample2() - 1) - 1) / 2.0);
      file << "\nMean:\t";
      if(programParameters->getMED() == 1) file << meanWithinGroup2->getBetaOrSigmaM() << "\t" << meanWithinGroup2->getSigmaE() << "\t" << meanWithinGroup2->getD1() << "\t***\t" << meanWithinGroup2Var << "\t" << meanWithinGroup2Skew << "\t" << meanWithinGroup2Kurt << endl;
      if(programParameters->getMED() == 2) file << meanWithinGroup2->getBetaOrSigmaM() << "\t" << meanWithinGroup2->getSigmaE() << "\t" << meanWithinGroup2->getD1() << "\t" << meanWithinGroup2->getD2() << "\t***\t" << meanWithinGroup2Var << "\t" << meanWithinGroup2Skew << "\t" << meanWithinGroup2Kurt << endl;
    }
    cout << "\n----------------------------------------------------------------" << endl;
    cout << "Estimate model parameters between groups " << programParameters->getSample1() << " and " << programParameters->getSample2() << endl;
    cout << "----------------------------------------------------------------" << endl;
    file << "\nEstimate model parameters between groups " << programParameters->getSample1() << " and " << programParameters->getSample2() << endl;
    for(long i = 0; i < data->getSampleSize(programParameters->getSample1() - 1); i++){
      for(long j = 0; j < data->getSampleSize(programParameters->getSample2() - 1); j++){
        cout << "\nComparisons between sample " << i + 1 << " (Group " << programParameters->getSample1() << ") and sample " << j + 1 << " (Group " << programParameters->getSample2() << "):" << endl;
        estimateFromTwoSamples(data,targetValues,targetDistribution2Samples,targetHistogramPlot2Samples,data->getSampleStart(programParameters->getSample1() - 1) + i,data->getSampleStart(programParameters->getSample2() - 1) + j);
        file << data->getSampleStart(programParameters->getSample1() - 1) + i + 1 << " - " << data->getSampleStart(programParameters->getSample2() - 1) + j + 1 << "\t";
        if(programParameters->getNonMutationalEffects() == 2){

          if(programParameters->getMethod() == 1){ 
            optimizer->getEstimate()->toString(programParameters);
            if(optimizer->getEstimate()->getQualityValue() == 1.0){
              if(programParameters->getMED() == 1) file << optimizer->getEstimate()->getBetaOrSigmaM() << "\t" << optimizer->getEstimate()->getSigmaE() << "\t" << optimizer->getEstimate()->getD1();
              if(programParameters->getMED() == 2) file << optimizer->getEstimate()->getBetaOrSigmaM() << "\t" << optimizer->getEstimate()->getSigmaE() << "\t" << optimizer->getEstimate()->getD1() << "\t" << optimizer->getEstimate()->getD2();
            } else {
              file << "Limit reached. No solution!";
            }
            meanBetweenGroups->addUpValues(optimizer->getEstimate());
          }
          if(programParameters->getMethod() == 2){ 
            // ML method only for the M&E-normal model
            mlOptimizer->getEstimate()->toString(programParameters);
            if(mlOptimizer->getEstimate()->getQualityValue() == 1.0){
              file << mlOptimizer->getEstimate()->getBetaOrSigmaM() << "\t" << mlOptimizer->getEstimate()->getSigmaE() << "\t" << mlOptimizer->getEstimate()->getD1();
            } else {
              file << "Limit reached. No solution!";
            }
            meanBetweenGroups->addUpValues(mlOptimizer->getEstimate());
          }

        } else {
          analyticEstimator->getEstimate()->toString(programParameters);
          if(programParameters->getMED() == 1) file << analyticEstimator->getEstimate()->getBetaOrSigmaM() << "\t" << analyticEstimator->getEstimate()->getSigmaE() << "\t" << analyticEstimator->getEstimate()->getD1();
          if(programParameters->getMED() == 2) file << analyticEstimator->getEstimate()->getBetaOrSigmaM() << "\t" << analyticEstimator->getEstimate()->getSigmaE() << "\t" << analyticEstimator->getEstimate()->getD1() << "\t" << analyticEstimator->getEstimate()->getD2();
          meanBetweenGroups->addUpValues(analyticEstimator->getEstimate());
        }
        file << "\t***\t" << targetValues->getVariance12() << "\t" << targetValues->getSkewness12() << "\t" << targetValues->getKurtosis12() << endl;
        meanBetweenGroupsVar = meanBetweenGroupsVar + targetValues->getVariance12();
        meanBetweenGroupsSkew = meanBetweenGroupsSkew + targetValues->getSkewness12();
        meanBetweenGroupsKurt = meanBetweenGroupsKurt + targetValues->getKurtosis12();
      }
    }
    meanBetweenGroups->divideValues(data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample2() - 1));
    meanBetweenGroupsVar = meanBetweenGroupsVar / (data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample2() - 1));
    meanBetweenGroupsSkew = meanBetweenGroupsSkew / (data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample2() - 1));
    meanBetweenGroupsKurt = meanBetweenGroupsKurt / (data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample2() - 1));
    file << "\nMean:\t";
    if(programParameters->getMED() == 1) file << meanBetweenGroups->getBetaOrSigmaM() << "\t" << meanBetweenGroups->getSigmaE() << "\t" << meanBetweenGroups->getD1() << "\t***\t" << meanBetweenGroupsVar << "\t" << meanBetweenGroupsSkew << "\t" << meanBetweenGroupsKurt << endl;
    if(programParameters->getMED() == 2) file << meanBetweenGroups->getBetaOrSigmaM() << "\t" << meanBetweenGroups->getSigmaE() << "\t" << meanBetweenGroups->getD1() << "\t" << meanBetweenGroups->getD2()  << "\t***\t" << meanBetweenGroupsVar << "\t" << meanBetweenGroupsSkew << "\t" << meanBetweenGroupsKurt << endl;

    file.close();

    delete meanWithinGroup1;
    delete meanWithinGroup2;
    delete meanBetweenGroups;
  }

  delete targetDistribution_1_2;
  delete targetHistogramPlot_1_2;
  delete targetDistribution_1_3;
  delete targetHistogramPlot_1_3;
  delete targetDistribution_2_3;
  delete targetHistogramPlot_2_3;
  delete targetDistribution2Samples;
  delete targetHistogramPlot2Samples;
}

// estimateFromTwoSamples:
// Estimates model parameters for the expression difference between a pair of individuals
// (called by performEstimationOnRealData in for analysisType == 2)
void Controller::estimateFromTwoSamples(Data* data,DistributionMeasures* targetValues,double* targetDistribution,double* targetHistogramPlot,long firstSample,long secondSample){

  // Initialisation
  MathFunctions mathFunctions = MathFunctions();
  double ave;
  double adev;
  double sdev;
  double var;
  double skew;
  double curt;

  // Calculate target distribution
  GraphicTools gs = GraphicTools();
  long indexBuffer;
  double lowerBound = 99999999;
  double upperBound = -99999999;
  for(long i = 0; i < data->getNumberOfGenes(); i++){
    targetDistribution[i] = data->getGene(firstSample,i) - data->getGene(secondSample,i);
  }
  gs.calcHistogramPlot(targetDistribution,data->getNumberOfGenes(),targetHistogramPlot,programParameters->getNumberOfBins(),&lowerBound,&upperBound);

  mathFunctions.calcMoments(targetDistribution,data->getNumberOfGenes(),&ave,&adev,&sdev,&var,&skew,&curt);
  targetValues->setVariance12(var);
  targetValues->setSkewness12(skew);
  targetValues->setKurtosis12(curt);
  if(programParameters->getNonMutationalEffects() == 2){
    // Complex simulation or outgroups not supported here (in each case only two samples are compared)
    if(programParameters->getMED() == 1){
      if(programParameters->getMethod() == 1){ 
        optimizer->estimateMENormalParametersForPairs(programParameters,threeTree,targetHistogramPlot,lowerBound,upperBound,targetValues,rnd,gammarnd);
      }
      if(programParameters->getMethod() == 2){
        mlOptimizer->estimateModelParametersMENormalPair(programParameters,data,rnd,gammarnd,firstSample,secondSample);
      }
    }
    if(programParameters->getMED() == 2){
      optimizer->estimateMEExtremeParametersForPairs(programParameters,threeTree,targetHistogramPlot,lowerBound,upperBound,targetValues,rnd,gammarnd);
    }
    if(programParameters->getMED() == 3){
      cout << "Gamma-distribution not supported in this mode" << endl;
    }
    cout << "\n----------------------------------------------------------------" << endl;
    cout << "\nEstimate:" << endl;
    if(optimizer->getEstimate()->getQualityValue() != 1.0){
      cout << "Limit reached. No solution!" << endl;
    }
  } else {
    analyticEstimator->calculateEstimate(programParameters,targetValues);
  }
}

// performBootstrappingSeries:
// Implements a bootstrapping and estimates model parameters on the bootstrap data
void Controller::performBootstrappingSeries(char* outputFile){

  // Initializing
  GraphicTools gs = GraphicTools();
  cout << "\nNumber of genes in table: " << data->getNumberOfGenes() << endl;
  DistributionMeasures* targetValues = new DistributionMeasures();
  Data* bootstrapData = new Data();
  bootstrapData->init(data->getNumberOfIndividuals(),data->getNumberOfGenes(),data->getNumberOfSamples());
  double* targetDistribution_1_2 = new double[data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample2() - 1) * data->getNumberOfGenes()];
  double* targetHistogramPlot_1_2 = new double[programParameters->getNumberOfBins()];
  double* targetDistribution_1_3 = new double[data->getSampleSize(programParameters->getSample1() - 1) * data->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes()];
  double* targetHistogramPlot_1_3 = new double[programParameters->getNumberOfBins()];
  double* targetDistribution_2_3 = new double[data->getSampleSize(programParameters->getSample2() - 1) * data->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes()];
  double* targetHistogramPlot_2_3 = new double[programParameters->getNumberOfBins()];
  double lowerBound_1_2 = 99999999;
  double upperBound_1_2 = -99999999;
  double lowerBound_1_3 = 99999999;
  double upperBound_1_3 = -99999999;
  double lowerBound_2_3 = 99999999;
  double upperBound_2_3 = -99999999;
  long indexBuffer;

  // Calculation of the side and the mutation effect skew
  data->calcDistributionMeasuresFromMeanOfAllPairs(programParameters,targetValues,true);
  cout << "\nOriginal data" << endl;
  targetValues->toString(programParameters);

  for(long runNumber = 0; runNumber < programParameters->getNumberOfRuns(); runNumber++){

    cout << "\nOptimizer run " << runNumber + 1 << "/" << programParameters->getNumberOfRuns() << endl;

    // Perform a bootstrap and present the current target values
    bootstrapData->createBootstrapData(data,rnd);
    bootstrapData->calcDistributionMeasuresFromMeanOfAllPairs(programParameters,targetValues,true);
    cout << "\nTarget values from bootstrapping:" << endl;
    targetValues->toString(programParameters);

    // Calculate target distribution on bootstrap data
    if(programParameters->getNonMutationalEffects() == 2){
      cout << "Calculate target distributions" << endl;
      bootstrapData->calcTargetDistribution(targetDistribution_1_2,programParameters->getSample1() - 1,programParameters->getSample2() - 1);
      if(programParameters->getSample3() != 0) bootstrapData->calcTargetDistribution(targetDistribution_1_3,programParameters->getSample1() - 1,programParameters->getSample3() - 1);
      if(programParameters->getSample3() != 0) bootstrapData->calcTargetDistribution(targetDistribution_2_3,programParameters->getSample2() - 1,programParameters->getSample3() - 1);
      gs.calcHistogramPlot(targetDistribution_1_2,bootstrapData->getSampleSize(programParameters->getSample1() - 1) * bootstrapData->getSampleSize(programParameters->getSample2() - 1) * data->getNumberOfGenes(),targetHistogramPlot_1_2,programParameters->getNumberOfBins(),&lowerBound_1_2,&upperBound_1_2);
      if(programParameters->getSample3() != 0) gs.calcHistogramPlot(targetDistribution_1_3,bootstrapData->getSampleSize(programParameters->getSample1() - 1) * bootstrapData->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes(),targetHistogramPlot_1_3,programParameters->getNumberOfBins(),&lowerBound_1_3,&upperBound_1_3);
      if(programParameters->getSample3() != 0) gs.calcHistogramPlot(targetDistribution_2_3,bootstrapData->getSampleSize(programParameters->getSample2() - 1) * bootstrapData->getSampleSize(programParameters->getSample3() - 1) * data->getNumberOfGenes(),targetHistogramPlot_2_3,programParameters->getNumberOfBins(),&lowerBound_2_3,&upperBound_2_3);
    }

    if(programParameters->getMED() == 3 || programParameters->getNonMutationalEffects() == 2){
      // Model which needs optimisation for parameter estimation
      if(programParameters->getNonMutationalEffects() == 1){
        // M-gamma model
        optimizer->estimateMGammaParameters(programParameters,bootstrapData,threeTree,targetValues,rnd,gammarnd);
      }
      if(programParameters->getNonMutationalEffects() == 2){
        if(programParameters->getMED() == 1){
          // M&E-normal model
          if(programParameters->getMethod() == 1) optimizer->estimateMENormalParameters(programParameters,bootstrapData,threeTree,targetHistogramPlot_1_2,lowerBound_1_2,upperBound_1_2,targetValues,rnd,gammarnd); // Chi^2
          if(programParameters->getMethod() == 2) mlOptimizer->estimateModelParametersMENormal(programParameters,bootstrapData,rnd,gammarnd); // ML
        }
        if(programParameters->getMED() == 2){
          // M&E-extreme model
          optimizer->estimateMEExtremeParameters(programParameters,bootstrapData,threeTree,targetHistogramPlot_1_2,lowerBound_1_2,upperBound_1_2,targetValues,rnd,gammarnd);
        }
      }
      cout << "\n----------------------------------------------------------------" << endl;
      cout << "\nResult:" << endl;
      // Get the results and store them in the result collector
      if(programParameters->getMethod() == 1 || programParameters->getMED() == 3){
        optimizer->getEstimate()->toString(programParameters);
        if(optimizer->getEstimate()->getQualityValue() != 1.0){
          cout << "Limit reached. No solution!" << endl;
        }
        resultCollector->setModelParameters(runNumber,*optimizer->getEstimate(),*targetValues);
        resultCollector->saveModelParameters(runNumber,outputFile);
      }
      if(programParameters->getMethod() == 2 && programParameters->getMED() != 3){
        mlOptimizer->getEstimate()->toString(programParameters);
        if(mlOptimizer->getEstimate()->getQualityValue() != 1.0){
          cout << "Limit reached. No solution!" << endl;
        }
        resultCollector->setModelParameters(runNumber,*mlOptimizer->getEstimate(),*targetValues);
        resultCollector->saveModelParameters(runNumber,outputFile);
      }
    } else {
      // Model whose parameters can be estimated analytically
      analyticEstimator->calculateEstimate(programParameters,targetValues);
      cout << "\nEstimate:" << endl;
      analyticEstimator->getEstimate()->toString(programParameters);
      // Get the results and store them in the result collector
      resultCollector->setModelParameters(runNumber,*analyticEstimator->getEstimate(),*targetValues);
    }
    if(programParameters->getOutputMode() > 1){
      cout << "\n----------------------------------------------------------------" << endl;
    }
  }

  // Save results
  cout << "\nSave table in file " << outputFile << endl;
  resultCollector->saveCompleteParameterTable(outputFile,false);

  delete bootstrapData;
  delete targetDistribution_1_2;
  delete targetHistogramPlot_1_2;
  delete targetDistribution_1_3;
  delete targetHistogramPlot_1_3;
  delete targetDistribution_2_3;
  delete targetHistogramPlot_2_3;
}

// performGenerateEstimateSeries:
// Generates data and estimates model parameters on the generated data
void Controller::performGenerateEstimateSeries(char* outputFile){

  // Initializing
  GraphicTools gs = GraphicTools();
  DistributionMeasures* targetValues = new DistributionMeasures();
  Data* artificialData = new Data();
  artificialData->init(3,programParameters->getNumberOfGenes(),3);
  artificialData->setSampleSize(0,1);
  artificialData->setSampleStart(0,0);
  artificialData->setSampleSize(1,1);
  artificialData->setSampleStart(1,1);
  artificialData->setSampleSize(2,1);
  artificialData->setSampleStart(2,2);

  double* targetDistribution_1_2 = new double[programParameters->getNumberOfSimulations()];
  double* targetHistogramPlot_1_2 = new double[programParameters->getNumberOfBins()];
  double* targetDistribution_1_3 = new double[programParameters->getNumberOfSimulations()];
  double* targetHistogramPlot_1_3 = new double[programParameters->getNumberOfBins()];
  double* targetDistribution_2_3 = new double[programParameters->getNumberOfSimulations()];
  double* targetHistogramPlot_2_3 = new double[programParameters->getNumberOfBins()];
  double lowerBound_1_2;
  double upperBound_1_2;
  double lowerBound_1_3;
  double upperBound_1_3;
  double lowerBound_2_3;
  double upperBound_2_3;
  long indexBuffer;

  // Set sample labels to 1, 2, 3 for data generation
  programParameters->setSample1(1);
  programParameters->setSample2(2);
  if(programParameters->getSample3() != 0) programParameters->setSample3(3);

  for(long runNumber = 0; runNumber < programParameters->getNumberOfRuns(); runNumber++){

    cout << "\nOptimizer run " << runNumber + 1 << "/" << programParameters->getNumberOfRuns() << endl;

    // Create artificial data and from given programParameters and calculate the target values
    threeTree->generation(programParameters,modelParameters,artificialData,rnd,gammarnd);

    // Calculate target distributions
    if(programParameters->getNonMutationalEffects() == 2){
      cout << "Calculate target distributions" << endl;
      artificialData->calcTargetDistribution(targetDistribution_1_2,programParameters->getSample1() - 1,programParameters->getSample2() - 1);
      if(programParameters->getSample3() != 0) artificialData->calcTargetDistribution(targetDistribution_1_3,programParameters->getSample1() - 1,programParameters->getSample3() - 1);
      if(programParameters->getSample3() != 0) artificialData->calcTargetDistribution(targetDistribution_2_3,programParameters->getSample2() - 1,programParameters->getSample3() - 1);
    }

    cout << "Target values from new created data:" << endl;
    artificialData->calcDistributionMeasuresFromMeanOfAllPairs(programParameters,targetValues,true);
    targetValues->toString(programParameters);

    lowerBound_1_2 = 99999999;
    upperBound_1_2 = -99999999;
    lowerBound_1_3 = 99999999;
    upperBound_1_3 = -99999999;
    lowerBound_2_3 = 99999999;
    upperBound_2_3 = -99999999;
    gs.calcHistogramPlot(targetDistribution_1_2,artificialData->getSampleSize(programParameters->getSample1() - 1) * artificialData->getSampleSize(programParameters->getSample2() - 1) * artificialData->getNumberOfGenes(),targetHistogramPlot_1_2,programParameters->getNumberOfBins(),&lowerBound_1_2,&upperBound_1_2);
    if(programParameters->getSample3() != 0) gs.calcHistogramPlot(targetDistribution_1_3,artificialData->getSampleSize(programParameters->getSample1() - 1) * artificialData->getSampleSize(programParameters->getSample3() - 1) * artificialData->getNumberOfGenes(),targetHistogramPlot_1_3,programParameters->getNumberOfBins(),&lowerBound_1_3,&upperBound_1_3);
    if(programParameters->getSample3() != 0) gs.calcHistogramPlot(targetDistribution_2_3,artificialData->getSampleSize(programParameters->getSample2() - 1) * artificialData->getSampleSize(programParameters->getSample3() - 1) * artificialData->getNumberOfGenes(),targetHistogramPlot_2_3,programParameters->getNumberOfBins(),&lowerBound_2_3,&upperBound_2_3);

    if(programParameters->getMED() == 3 || programParameters->getNonMutationalEffects() == 2){
      // Model which needs optimisation for parameter estimation
      if(programParameters->getNonMutationalEffects() == 1){
        // M-gamma model
        optimizer->estimateMGammaParameters(programParameters,artificialData,threeTree,targetValues,rnd,gammarnd);
      }
      if(programParameters->getNonMutationalEffects() == 2){
        if(programParameters->getMED() == 1){
          // M-normal model
          if(programParameters->getMethod() == 1)  optimizer->estimateMENormalParameters(programParameters,artificialData,threeTree,targetHistogramPlot_1_2,lowerBound_1_2,upperBound_1_2,targetValues,rnd,gammarnd); // Chi^2
          if(programParameters->getMethod() == 2) mlOptimizer->estimateModelParametersMENormal(programParameters,artificialData,rnd,gammarnd); // ML
        }
        if(programParameters->getMED() == 2){
          // M-extreme model
          optimizer->estimateMEExtremeParameters(programParameters,artificialData,threeTree,targetHistogramPlot_1_2,lowerBound_1_2,upperBound_1_2,targetValues,rnd,gammarnd);
        }
      }
      cout << "\n----------------------------------------------------------------" << endl;
      cout << "\nResult:" << endl;
      // Get the results and store them in the result collector
      if(programParameters->getMethod() == 1 || programParameters->getMED() == 3){
        optimizer->getEstimate()->toString(programParameters);
        if(optimizer->getEstimate()->getQualityValue() != 1.0){
          cout << "Limit reached. No solution!" << endl;
        }
        resultCollector->setModelParameters(runNumber,*optimizer->getEstimate(),*targetValues);
        resultCollector->saveModelParameters(runNumber,outputFile);
      }
      if(programParameters->getMethod() == 2 && programParameters->getMED() != 3){
        mlOptimizer->getEstimate()->toString(programParameters);
        if(mlOptimizer->getEstimate()->getQualityValue() != 1.0){
          cout << "Limit reached. No solution!" << endl;
        }
        resultCollector->setModelParameters(runNumber,*mlOptimizer->getEstimate(),*targetValues);
        resultCollector->saveModelParameters(runNumber,outputFile);
      }
    } else {
      // Model whose parameters can be estimated analytically
      analyticEstimator->calculateEstimate(programParameters,targetValues);
      cout << "\nEstimate:" << endl;
      analyticEstimator->getEstimate()->toString(programParameters);
      // Get the results and store them in the result collector
      resultCollector->setModelParameters(runNumber,*analyticEstimator->getEstimate(),*targetValues);
    }
    if(programParameters->getOutputMode() > 1){
      cout << "\n----------------------------------------------------------------" << endl;
    }
  }

  // Save results
  cout << "\nSave table in file " << outputFile << endl;
  resultCollector->saveCompleteParameterTable(outputFile,false);

  delete artificialData;
  delete targetDistribution_1_2;
  delete targetHistogramPlot_1_2;
  delete targetDistribution_1_3;
  delete targetHistogramPlot_1_3;
  delete targetDistribution_2_3;
  delete targetHistogramPlot_2_3;
}
