/***************************************************************************
                          data.cpp  -  description
                             -------------------
    begin                : Thu Feb 17 2005
    copyright            : (C) 2005 by Michael Rosskopf
    email                : michael@bioinf.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "data.h"

// Constructors and destructors
Data::Data(){
  numberOfSamples = 0;
  numberOfIndividuals = 0;
  numberOfGenes = 0;
  genes = 0;
  sampleSize = 0;
  sampleStart = 0;
  labels = 0;
  probeSetID = 0;
}
Data::~Data(){
  if(genes != 0) delete genes;
  if(sampleSize != 0) delete sampleSize;
  if(sampleStart != 0) delete sampleStart;
  if(labels != 0) delete labels;
  if(probeSetID != 0){
    for(long i = 0; i < numberOfGenes; i++){
      if(probeSetID[i] != 0) delete probeSetID[i];
    }
    delete probeSetID;
  }
}

// init
void Data::init(long _numberOfIndividuals,long _numberOfGenes){
  numberOfSamples = 1;
  numberOfIndividuals = _numberOfIndividuals;
  numberOfGenes = _numberOfGenes;
  labels = new long[numberOfIndividuals];
  genes = new double[numberOfIndividuals * numberOfGenes];
  probeSetID = new char*[numberOfGenes];
}
void Data::init(long _numberOfIndividuals,long _numberOfGenes,long _numberOfSamples){
  numberOfSamples = _numberOfSamples;
  numberOfIndividuals = _numberOfIndividuals;
  numberOfGenes = _numberOfGenes;
  labels = new long[numberOfIndividuals];
  sampleSize = new long[numberOfSamples];
  sampleStart = new long[numberOfSamples];
  genes = new double[numberOfIndividuals * numberOfGenes];
  probeSetID = new char*[numberOfGenes];
}

// get- and set-methods
long Data::getNumberOfSamples(){
  return numberOfSamples;
}
long Data::getNumberOfIndividuals(){
  return numberOfIndividuals;
}
long Data::getNumberOfGenes(){
  return numberOfGenes;
}
long Data::getLabel(long _sample){
  return labels[_sample];
}
long Data::getSampleSize(long _sampleNumber){
  return sampleSize[_sampleNumber];
}
long Data::getSampleStart(long _sampleNumber){
  return sampleStart[_sampleNumber];
}
char* Data::getProbeSetID(long index){
  return probeSetID[index];
}
double Data::getGene(long _sample,long _gene){
  return genes[(numberOfIndividuals * _gene) + _sample];
}
void Data::setLabel(long index,long _label){
  labels[index] = _label;
}
void Data::setSampleSize(long _sampleNumber,long _value){
  sampleSize[_sampleNumber] = _value;
}
void Data::setSampleStart(long _sampleNumber,long _value){
  sampleStart[_sampleNumber] = _value;
}
void Data::setGene(long _sample,long _gene,double expressionValue){
  genes[(numberOfIndividuals * _gene) + _sample] = expressionValue;
}

// toString
void Data::toString(){
  cout << "ProbeSetID\t";
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    cout << labels[i] << "\t";
  }
  cout << labels[numberOfIndividuals - 1] << "\n";
  for(long i = 0; i < numberOfGenes; i++){
  cout << i + 1 << "\t";
    for(long j = 0; j < numberOfIndividuals - 1; j++){
      cout << genes[(numberOfIndividuals * i) + j] << "\t";
    }
    cout << genes[(numberOfIndividuals * i) + numberOfIndividuals - 1] << "\n";
  }
}

// dataCheck: Checks whether sample labels are in the data set
bool Data::dataCheck(ProgramParameters* programParameters){
  bool globalvalid = true;
  bool valid;
  // Checks whether sample1 is inside the data set
  valid = false;
  for(long i = 0; i < numberOfIndividuals; i++){
    if(programParameters->getSample1() == labels[i]) valid = true;
  }
  if(!valid){
    cout << "\nError: Data set does not contain sample1 = " << programParameters->getSample1() << "." << endl;
    globalvalid = false;
  }
  // Checks whether sample2 is inside the data set
  valid = false;
  for(long i = 0; i < numberOfIndividuals; i++){
    if(programParameters->getSample2() == labels[i]) valid = true;
  }
  if(!valid){
    cout << "\nError: Data set does not contain sample2 = " << programParameters->getSample2() << "." << endl;
    globalvalid = false;
  }
  // Checks whether sample3 is inside the data set
  if(programParameters->getSample3() != 0){
    valid = false;
    for(long i = 0; i < numberOfIndividuals; i++){
      if(programParameters->getSample3() == labels[i]) valid = true;
    }
    if(!valid){
      cout << "\nError: Data set does not contain sample3 = " << programParameters->getSample3() << "." << endl;
      globalvalid = false;
    }
  }
  return globalvalid;
}

// calcVarianceWithinASample: 
// Calculates the gene expression variance within a sample
// (called by the calcDistributionMeasuresFromMeanOfAllPairs method)
// The results a written to targetValues.
void Data::calcVarianceWithinASample(ProgramParameters* programParameters,DistributionMeasures* targetValues){

  // Initialisation
  MathFunctions mathFunctions = MathFunctions();
  double average;
  double averageDeviation;
  double standardDeviation;
  double variance;
  double skewness;
  double kurtosis;
  double* differenceVector = new double[numberOfGenes];
  double varianceWithinGroup;
  long i,j,k; // Counters

  // Calculate variance within group 1, if it contains more than one individual
  if(sampleSize[programParameters->getSample1() - 1] > 1){
    varianceWithinGroup = 0.0;
    for(i = 0; i < sampleSize[programParameters->getSample1() - 1] - 1; i++){
      for(j = i + 1; j < sampleSize[programParameters->getSample1() - 1]; j++){
        for(k = 0; k < numberOfGenes; k++){
          differenceVector[k] = genes[(numberOfIndividuals * k) + sampleStart[programParameters->getSample1() - 1] + i] - genes[(numberOfIndividuals * k) + sampleStart[programParameters->getSample1() - 1] + j];
        }
        mathFunctions.calcMoments(differenceVector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
        varianceWithinGroup = varianceWithinGroup + variance;
      }
    }
    varianceWithinGroup = varianceWithinGroup / ((sampleSize[programParameters->getSample1() - 1] * (sampleSize[programParameters->getSample1() - 1] - 1)) / 2.0);
    targetValues->setVarianceWithin1(varianceWithinGroup);
  } else targetValues->setVarianceWithin1(0.0);

  // Calculate variance within group 2, if it contains more than one individual
  if(sampleSize[programParameters->getSample2() - 1] > 1){
    varianceWithinGroup = 0.0;
    for(i = 0; i < sampleSize[programParameters->getSample2() - 1] - 1; i++){
      for(j = i + 1; j < sampleSize[programParameters->getSample2() - 1]; j++){
        for(k = 0; k < numberOfGenes; k++){
          differenceVector[k] = genes[(numberOfIndividuals * k) + sampleStart[programParameters->getSample2() - 1] + i] - genes[(numberOfIndividuals * k) + sampleStart[programParameters->getSample2() - 1] + j];
        }
        mathFunctions.calcMoments(differenceVector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
        varianceWithinGroup = varianceWithinGroup + variance;
      }
    }
    varianceWithinGroup = varianceWithinGroup / ((sampleSize[programParameters->getSample2() - 1] * (sampleSize[programParameters->getSample2() - 1] - 1)) / 2.0);
    targetValues->setVarianceWithin2(varianceWithinGroup);
  } else targetValues->setVarianceWithin2(0.0);

  // Calculate variance within group 1, if it contains more than one individual
  if((programParameters->getSample3() != 0) && (sampleSize[programParameters->getSample3() - 1] > 1)){
    varianceWithinGroup = 0.0;
    for(i = 0; i < sampleSize[programParameters->getSample3() - 1] - 1; i++){
      for(j = i + 1; j < sampleSize[programParameters->getSample3() - 1]; j++){
        for(k = 0; k < numberOfGenes; k++){
          differenceVector[k] = genes[(numberOfIndividuals * k) + sampleStart[programParameters->getSample3() - 1] + i] - genes[(numberOfIndividuals * k) + sampleStart[programParameters->getSample3() - 1] + j];
        }
        mathFunctions.calcMoments(differenceVector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
        varianceWithinGroup = varianceWithinGroup + variance;
      }
    }
    varianceWithinGroup = varianceWithinGroup / ((sampleSize[programParameters->getSample3() - 1] * (sampleSize[programParameters->getSample3() - 1] - 1)) / 2.0);
    targetValues->setVarianceWithin3(varianceWithinGroup);
  } else targetValues->setVarianceWithin3(0.0);

  delete differenceVector;
}

// calcDistributionMeasuresFromMeanOfAllPairs:
// Calculates the mean moments of the gene expression differences between all pairs of two samples
// The results a stored in targetValues.
// Important for the M-gamma model: If sideAndMirrowUpdate == true, 
// the choice for the gamma distribution or its mirrored version is calculated.
void Data::calcDistributionMeasuresFromMeanOfAllPairs(ProgramParameters* programParameters,DistributionMeasures* targetValues,bool sideAndMirrowUpdate){

  // Initialisation
  MathFunctions mathFunctions = MathFunctions();
  double average;
  double averageDeviation;
  double standardDeviation;
  double variance;
  double skewness;
  double kurtosis;
  double median;
  double s1iSkewMean = 0.0;
  double s1iPearsonsSkewMean = 0.0;
  long s1iCounter;
  long overallS1iCounter = 0;
  double s2iSkewMean = 0.0;
  double s2iPearsonsSkewMean = 0.0;
  long s2iCounter;
  long overallS2iCounter = 0;
  double* differenceVector = new double[numberOfGenes];
  double* varianceMean = new double[numberOfSamples * (numberOfSamples - 1) / 2];
  double* skewnessMean = new double[numberOfSamples * (numberOfSamples - 1) / 2];
  double* kurtosisMean = new double[numberOfSamples * (numberOfSamples - 1) / 2];


  // Calculate variance, skewness and kurtosis
  for(long i = 0; i < numberOfSamples - 1; i++){
    for(long j = i + 1; j < numberOfSamples; j++){
      varianceMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = 0;
      skewnessMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = 0;
      kurtosisMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = 0;
      for(long k = 0; k < sampleSize[i]; k++){
        for(long l = 0; l < sampleSize[j]; l++){
          for(long m = 0; m < numberOfGenes; m++){
            differenceVector[m] = genes[(numberOfIndividuals * m) + sampleStart[i] + k] - genes[(numberOfIndividuals * m) + sampleStart[j] + l];
          }
          mathFunctions.calcMoments(differenceVector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
          varianceMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = varianceMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] + variance;
          skewnessMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = skewnessMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] + skewness;
          kurtosisMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = kurtosisMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] + kurtosis;
        }
      }
      varianceMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = varianceMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] / (sampleSize[i] * sampleSize[j]);
      skewnessMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = skewnessMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] / (sampleSize[i] * sampleSize[j]);
      kurtosisMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] = kurtosisMean[(((numberOfSamples - i - 1) * (numberOfSamples - i - 2)) / 2) + (numberOfSamples - j - 1)] / (sampleSize[i] * sampleSize[j]);
    }
  }

  if(programParameters->getSample3() != 0){

    // Calculates the sample-1-intermediate skewness
//    ofstream s1ifile("_s1i-genes.txt");
    for(long i = 0; i < sampleSize[programParameters->getSample1() - 1]; i++){
      for(long j = 0; j < sampleSize[programParameters->getSample2() - 1]; j++){
        for(long k = 0; k < sampleSize[programParameters->getSample3() - 1]; k++){
          s1iCounter = 0;
          for(long l = 0; l < numberOfGenes; l++){
            if(((genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k]) || (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j]))   ||   ((genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] == genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j] != genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i]))){
              differenceVector[s1iCounter] = genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] - genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j];
              s1iCounter++;
              overallS1iCounter++;
//              s1ifile << differenceVector[s1iCounter] << endl;
            }
          }
          mathFunctions.calcMoments(differenceVector,s1iCounter,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
          s1iSkewMean = s1iSkewMean + skewness;
          if(programParameters->getSkewType() == 2){
            mathFunctions.calcMedianRearrange(differenceVector,s1iCounter,&median);
            s1iPearsonsSkewMean = s1iPearsonsSkewMean + ((3.0 * (average - median)) / standardDeviation);
          }
        }
      }
    }
    s1iSkewMean = s1iSkewMean / (sampleSize[programParameters->getSample1() - 1] * sampleSize[programParameters->getSample2() - 1] * sampleSize[programParameters->getSample3() - 1]);
    s1iPearsonsSkewMean = s1iPearsonsSkewMean / (sampleSize[programParameters->getSample1() - 1] * sampleSize[programParameters->getSample2() - 1] * sampleSize[programParameters->getSample3() - 1]);
//    s1ifile.close();

    // Calculates the sample-2-intermediate skewness
//    ofstream s2ifile("_s2i-genes.txt");
    for(long i = 0; i < sampleSize[programParameters->getSample1() - 1]; i++){
      for(long j = 0; j < sampleSize[programParameters->getSample2() - 1]; j++){
        for(long k = 0; k < sampleSize[programParameters->getSample3() - 1]; k++){
          s2iCounter = 0;
          for(long l = 0; l < numberOfGenes; l++){
            if(((genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k]) || (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i]))   ||   ((genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j] == genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] != genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j]))){
              differenceVector[s2iCounter] = genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] - genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j];
              s2iCounter++;
              overallS2iCounter++;
//              s2ifile << differenceVector[s2iCounter] << endl;
            }
          }
          mathFunctions.calcMoments(differenceVector,s2iCounter,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
          s2iSkewMean = s2iSkewMean + skewness;
          if(programParameters->getSkewType() == 2){
            mathFunctions.calcMedianRearrange(differenceVector,s2iCounter,&median);
            s2iPearsonsSkewMean = s2iPearsonsSkewMean + ((3.0 * (average - median)) / standardDeviation);
          }
        }
      }
    }
    s2iSkewMean = s2iSkewMean / (sampleSize[programParameters->getSample1() - 1] * sampleSize[programParameters->getSample2() - 1] * sampleSize[programParameters->getSample3() - 1]);
    s2iPearsonsSkewMean = s2iPearsonsSkewMean / (sampleSize[programParameters->getSample1() - 1] * sampleSize[programParameters->getSample2() - 1] * sampleSize[programParameters->getSample3() - 1]);
//    s2ifile.close();

    // Calculates the sample-3-intermediate genes (only for s3ifile)
//    ofstream s3ifile("_s3i-genes.txt");
    for(long i = 0; i < sampleSize[programParameters->getSample1() - 1]; i++){
      for(long j = 0; j < sampleSize[programParameters->getSample2() - 1]; j++){
        for(long k = 0; k < sampleSize[programParameters->getSample3() - 1]; k++){
          for(long l = 0; l < numberOfGenes; l++){
            if(((genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j]) || (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k]) && (genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample3() - 1] + k] < genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i]))   ||   ((genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] == genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j]))){
//            s3ifile << genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample1() - 1] + i] - genes[(numberOfIndividuals * l) + sampleStart[programParameters->getSample2() - 1] + j] << endl;
            }
          }
        }
      }
    }
//    s3ifile.close();
  }

  targetValues->setVariance12(varianceMean[(((numberOfSamples - programParameters->getSample1()) * (numberOfSamples - programParameters->getSample1() - 1)) / 2) + (numberOfSamples - programParameters->getSample2())]);
  targetValues->setSkewness12(skewnessMean[(((numberOfSamples - programParameters->getSample1()) * (numberOfSamples - programParameters->getSample1() - 1)) / 2) + (numberOfSamples - programParameters->getSample2())]);
  targetValues->setKurtosis12(kurtosisMean[(((numberOfSamples - programParameters->getSample1()) * (numberOfSamples - programParameters->getSample1() - 1)) / 2) + (numberOfSamples - programParameters->getSample2())]);

  if(programParameters->getSample3() != 0){
    targetValues->setVariance13(varianceMean[(((numberOfSamples - programParameters->getSample1()) * (numberOfSamples - programParameters->getSample1() - 1)) / 2) + (numberOfSamples - programParameters->getSample3())]);
    targetValues->setVariance23(varianceMean[(((numberOfSamples - programParameters->getSample2()) * (numberOfSamples - programParameters->getSample2() - 1)) / 2) + (numberOfSamples - programParameters->getSample3())]);
    targetValues->setSkewness13(skewnessMean[(((numberOfSamples - programParameters->getSample1()) * (numberOfSamples - programParameters->getSample1() - 1)) / 2) + (numberOfSamples - programParameters->getSample3())]);
    targetValues->setSkewness23(skewnessMean[(((numberOfSamples - programParameters->getSample2()) * (numberOfSamples - programParameters->getSample2() - 1)) / 2) + (numberOfSamples - programParameters->getSample3())]);
    targetValues->setKurtosis13(kurtosisMean[(((numberOfSamples - programParameters->getSample1()) * (numberOfSamples - programParameters->getSample1() - 1)) / 2) + (numberOfSamples - programParameters->getSample3())]);
    targetValues->setKurtosis23(kurtosisMean[(((numberOfSamples - programParameters->getSample2()) * (numberOfSamples - programParameters->getSample2() - 1)) / 2) + (numberOfSamples - programParameters->getSample3())]);
    targetValues->setS1iSkew(s1iSkewMean);
    targetValues->setS2iSkew(s2iSkewMean);
    if(programParameters->getSkewType() == 2){
      targetValues->setS1iPearsonsSkew(s1iPearsonsSkewMean);
      targetValues->setS2iPearsonsSkew(s2iPearsonsSkewMean);
    } else {
      targetValues->setS1iPearsonsSkew(0.0);
      targetValues->setS2iPearsonsSkew(0.0);
    }
    targetValues->setOverallS1iCounter(overallS1iCounter);    
    targetValues->setOverallS2iCounter(overallS2iCounter);

    // Decide which characteristic (S1iSkew or S2iSkew is used for optimization)
    // and decide wheather a normal gamma distribution or a mirrored one is used
    if(sideAndMirrowUpdate){
//      cout << "-> Calculate side and mutation effect skewnessness" << endl;
      if(programParameters->getSkewType() == 1){
        if(targetValues->getS1iSkew() < targetValues->getS2iSkew()){
          // Normal gamma distribution
          targetValues->setMirrored(0);
          if(overallS1iCounter > overallS2iCounter){
            // S1iSkew-Distribution greater
            if(targetValues->getS1iSkew() < 0.0 && targetValues->getS2iSkew() < 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iSkew() < 0.0 && targetValues->getS2iSkew() > 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iSkew() > 0.0 && targetValues->getS2iSkew() > 0.0){
              targetValues->setSide(1);
            }
          } else {
            // S2iSkew-Distribution greater
            if(targetValues->getS1iSkew() < 0.0 && targetValues->getS2iSkew() < 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iSkew() < 0.0 && targetValues->getS2iSkew() > 0.0){
              targetValues->setSide(1);
            }
            if(targetValues->getS1iSkew() > 0.0 && targetValues->getS2iSkew() > 0.0){
              targetValues->setSide(1);
            }
          }
        } else {
          // Mirrored gamma distribution
          targetValues->setMirrored(1);
          if(overallS1iCounter > overallS2iCounter){
            // S1iSkew-Distribution greater
            if(targetValues->getS1iSkew() > 0.0 && targetValues->getS2iSkew() > 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iSkew() > 0.0 && targetValues->getS2iSkew() < 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iSkew() < 0.0 && targetValues->getS2iSkew() < 0.0){
              targetValues->setSide(1);
            }
          } else {
            // S2iSkew-Distribution greater
            if(targetValues->getS1iSkew() > 0.0 && targetValues->getS2iSkew() > 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iSkew() > 0.0 && targetValues->getS2iSkew() < 0.0){
              targetValues->setSide(1);
            }
            if(targetValues->getS1iSkew() < 0.0 && targetValues->getS2iSkew() < 0.0){
              targetValues->setSide(1);
            }
          }
        }
      }
      if(programParameters->getSkewType() == 2){
        if(targetValues->getS1iPearsonsSkew() < targetValues->getS2iPearsonsSkew()){
          // Normal gamma distribution
          targetValues->setMirrored(0);
          if(overallS1iCounter > overallS2iCounter){
            // S1iSkew-Distribution greater
            if(targetValues->getS1iPearsonsSkew() < 0.0 && targetValues->getS2iPearsonsSkew() < 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iPearsonsSkew() < 0.0 && targetValues->getS2iPearsonsSkew() > 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iPearsonsSkew() > 0.0 && targetValues->getS2iPearsonsSkew() > 0.0){
              targetValues->setSide(1);
            }
          } else {
            // S2iSkew-Distribution greater
            if(targetValues->getS1iPearsonsSkew() < 0.0 && targetValues->getS2iPearsonsSkew() < 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iPearsonsSkew() < 0.0 && targetValues->getS2iPearsonsSkew() > 0.0){
              targetValues->setSide(1);
            }
            if(targetValues->getS1iPearsonsSkew() > 0.0 && targetValues->getS2iPearsonsSkew() > 0.0){
              targetValues->setSide(1);
            }
          }
        } else {
          // Mirrored gamma distribution
          targetValues->setMirrored(1);
          if(overallS1iCounter > overallS2iCounter){
            // S1iSkew-Distribution greater
            if(targetValues->getS1iPearsonsSkew() > 0.0 && targetValues->getS2iPearsonsSkew() > 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iPearsonsSkew() > 0.0 && targetValues->getS2iPearsonsSkew() < 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iPearsonsSkew() < 0.0 && targetValues->getS2iPearsonsSkew() < 0.0){
              targetValues->setSide(1);
            }
          } else {
            // S2iSkew-Distribution greater
            if(targetValues->getS1iPearsonsSkew() > 0.0 && targetValues->getS2iPearsonsSkew() > 0.0){
              targetValues->setSide(0);
            }
            if(targetValues->getS1iPearsonsSkew() > 0.0 && targetValues->getS2iPearsonsSkew() < 0.0){
              targetValues->setSide(1);
            }
            if(targetValues->getS1iPearsonsSkew() < 0.0 && targetValues->getS2iPearsonsSkew() < 0.0){
              targetValues->setSide(1);
            }
          }
        }
      }
    }
  }

  delete kurtosisMean;
  delete skewnessMean;
  delete varianceMean;
  delete differenceVector;

  calcVarianceWithinASample(programParameters,targetValues);
}

// createBootstrapData: Creates a bootstrap dataset from data
void Data::createBootstrapData(Data* data,RandomNumberGenerator* rnd){

  // Calculate the bootstrap data
  long* bootstrapSampleList = new long[numberOfIndividuals];
  long* bootstrapGeneList = new long[numberOfGenes];
  for(long i = 0; i < data->getNumberOfSamples(); i++){
    for(long j = 0; j < data->getSampleSize(i); j++){
      // Over genes and samples
      bootstrapSampleList[data->getSampleStart(i) + j] = data->getSampleStart(i) + long(rnd->getUniformDistributedNumber(0,data->getSampleSize(i)));
      // Only over genes
    }
  }
  for(long i = 0; i < numberOfGenes; i++){
    bootstrapGeneList[i] = long(rnd->getUniformDistributedNumber(0,numberOfGenes));
  }

  // Copy the data along the calculated bootstrap indices
  for(long i = 0; i < numberOfSamples; i++){
    sampleSize[i] = data->getSampleSize(i);
    sampleStart[i] = data->getSampleStart(i);
  }
  for(long i = 0; i < data->getNumberOfIndividuals(); i++){
    setLabel(i,data->getLabel(i));
    for(long j = 0; j < data->getNumberOfGenes(); j++){
      setGene(i,j,data->getGene(bootstrapSampleList[i],bootstrapGeneList[j]));
    }
  }

  delete bootstrapSampleList;
  delete bootstrapGeneList;
}

// calcTargetDistribution:
// Calculates the difference distribution between two groups of the dataset.
// The distribution contains the differences between all pairs for all genes:
// #group1 * #group2 * #genes
void Data::calcTargetDistribution(double *targetDistribution,long sample1,long sample2){

  long index = 0;
  for(long i = 0; i < sampleSize[sample1]; i++){
    for(long j = 0; j < sampleSize[sample2]; j++){
      for(long k = 0; k < numberOfGenes; k++){
        targetDistribution[index++] = getGene(sampleStart[sample1] + i,k) - getGene(sampleStart[sample2] + j,k);
      }
    }
  }
}

// checkDataSize: Called from the load-method. This method is used
//                to check the data size and to initialize the data structure.
void Data::checkDataSize(char* filename){
  char charbuffer;
  numberOfIndividuals = 0;
  numberOfGenes = 0;
  ifstream file(filename);
  while(file.get(charbuffer)){
    if(charbuffer == '\n') break;
    if(charbuffer == '\t') numberOfIndividuals++;
  }
  while(file.get(charbuffer)){
    if(charbuffer == '\n') numberOfGenes++;
  }
  file.close();
  init(numberOfIndividuals,numberOfGenes);
}

// load: Loads a dataset. The initialization occurs automatically
void Data::load(char* filename){
  checkDataSize(filename);
  char charbuffer;
  char stringbuffer[64];
  ifstream file(filename);
  file.get(stringbuffer,64,'\t');
  file.get(charbuffer);
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    file.get(stringbuffer,64,'\t');
    labels[i] = atol(stringbuffer);
    file.get(charbuffer);
  }
  file.get(stringbuffer,64,'\n');
  labels[numberOfIndividuals - 1] = atol(stringbuffer);
  file.get(charbuffer);

  for(long i = 0; i < numberOfGenes; i++){
  file.get(stringbuffer,64,'\t');
  probeSetID[i] = new char[sizeof(stringbuffer)];
  strcpy(probeSetID[i],stringbuffer);
  file.get(charbuffer);
    for(long j = 0; j < numberOfIndividuals - 1; j++){
      file.get(stringbuffer,64,'\t');
      genes[(numberOfIndividuals * i) + j] = atof(stringbuffer);
      file.get(charbuffer);
    }
    file.get(stringbuffer,64,'\n');
    genes[(numberOfIndividuals * i) + numberOfIndividuals - 1] = atof(stringbuffer);
    file.get(charbuffer);
  }
  file.close();
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    if(labels[i] != labels[i + 1]) numberOfSamples++;
  }
  sampleSize = new long[numberOfSamples];
  sampleStart = new long[numberOfSamples];
  for(long i = 0; i < numberOfSamples; i ++){
    sampleSize[i] = 0;
    sampleStart[i] = 0;
  }
  long groupCounter = 1;
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    if(labels[i] != labels[i + 1]){
      sampleStart[groupCounter] = i + 1;
      groupCounter++;
    }
    sampleSize[labels[i] - 1]++;
  }
  sampleSize[labels[numberOfIndividuals - 1] - 1]++;
}

// save: Saves a dataset.
void Data::save(char* filename){
  ofstream file(filename);
  file.precision(10);
  file << "ProbeSetID\t";
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    file << labels[i] << "\t";
  }
  file << labels[numberOfIndividuals - 1] << "\n";
  for(long i = 0; i < numberOfGenes; i++){
    file << i + 1 << "\t";
    for(long j = 0; j < numberOfIndividuals - 1; j++){
      file << genes[(numberOfIndividuals * i) + j] << "\t";
    }
    file << genes[(numberOfIndividuals * i) + numberOfIndividuals - 1] << "\n";
  }
  file.close();
}
