/***************************************************************************
 *   Copyright (C) 2006 by Michael Rosskopf                                *
 *   rosskopf@cs.uni-duesseldorf.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "mloptimizer.h"

// Constructors and destructors
MLOptimizer::MLOptimizer(){
  ms = new MathFunctions();
  estimate = new ModelParameters();
  factorialValue = new double[21];
  for(int i = 0; i <= 20; i++){
    factorialValue[i] = factorial(i);
  }
}
MLOptimizer::~MLOptimizer(){
  if(estimate != 0) delete estimate;
  if(factorialValue != 0) delete factorialValue;
  if(ms != 0) delete ms;
}

// Returns the middle estimate which is the best one
ModelParameters* MLOptimizer::getEstimate(){
  return estimate;
}

// factorial: Calculates the factorial of "value"
double MLOptimizer::factorial(long value){
  double factorial = 1;
  for(long i = 1; i <= value; i++){
    factorial = factorial * i;
  }
  return factorial;
}

// poissonPDF: Calculates the frequency of "k" events for a Poisson distribution with parameter "rate"
double MLOptimizer::poissonPDF(double rate,long k){
  return (pow(2.718281828,-rate) * pow(rate,k)) / factorialValue[k];
}

// normalPDF: Calculates the density of a normal distribution
double MLOptimizer::normalPDF(double standardDeviation,double x){
  return (1.0 / (pow(2.0 * PI,0.5) * standardDeviation)) * pow(2.718281828,-pow(x,2.0) / (2.0 * pow(standardDeviation,2.0)));
}

// compoundPoissonMENormalPDF:
// Calculates the density of a compound Poisson process with normal distributed random variables.
// Parameters: Rate of the Poisson process, standard deviation of the normal distributed events
// max gives the maximal number of iterations for the numerical solution (it is infinite in reality)
double MLOptimizer::compoundPoissonMENormalPDF(double rate,double stdMut,double stdEnv,double x,long max){
  double result = 0.0;
  double buffer = 0.0;
  for(long i = 0; i <= max; i++){
    buffer = poissonPDF(rate,i) * normalPDF(pow(pow(stdMut,2.0) * i + 2.0 * pow(stdEnv,2.0),0.5),x);
    if(!isinf(buffer) && !isnan(buffer) && (buffer > 0.0)) result = result + buffer;
  }
  return result;
}

// calcLikelihoodMENormal:
// Calculates the likelihood for the parameters of the M&E-normal model given data
double MLOptimizer::calcLikelihoodMENormal(double stdMut,double stdEnv, double d,ProgramParameters* programParameters,Data* data){

  long sizeSample1 = data->getSampleSize(programParameters->getSample1() - 1);
  long sizeSample2 = data->getSampleSize(programParameters->getSample2() - 1);
  long numberOfGenes = data->getNumberOfGenes();
  double expressionDifference;
  double buffer = 0.0;
  long i,j,k;  
  for(i = 0; i < sizeSample1; i++){
    for(j = 0; j < sizeSample2; j++){
      for(k = 0; k < numberOfGenes; k++){
        expressionDifference = data->getGene(data->getSampleStart(programParameters->getSample1() - 1) + i,k) - data->getGene(data->getSampleStart(programParameters->getSample2() - 1) + j,k);
        buffer = buffer + (-log(compoundPoissonMENormalPDF(d,stdMut,stdEnv,expressionDifference,10)));
      }
    }
  }
  return buffer;
}

// calcLikelihoodMENormalPair:
// Calculates the likelihood for the parameters of the M&E-normal model given a pair of individuals from the data
double MLOptimizer::calcLikelihoodMENormalPair(double stdMut,double stdEnv, double d,ProgramParameters* programParameters,Data* data,int individual1,int individual2){

  long numberOfGenes = data->getNumberOfGenes();
  double expressionDifference;
  double buffer = 0.0;
  long i,j,k;  
  for(k = 0; k < numberOfGenes; k++){
    expressionDifference = data->getGene(individual1,k) - data->getGene(individual2,k);
    buffer = buffer + (-log(compoundPoissonMENormalPDF(d,stdMut,stdEnv,expressionDifference,10)));
  }
  return buffer;
}

// scaleMENormal:
// Implements reflections of the Downhill Simplex Method
// (called by estimateModelParametersNormalMutNormalEnv)
double MLOptimizer::scaleMENormal(ModelParameters **searchPoint,ModelParameters *buffer,ModelParameters *psum,int worst,double factor,ProgramParameters *programParameters,Data *data,RandomNumberGenerator *rnd,Random *gammarnd){

  double factor1;
  double factor2;
  double rating;

  factor1 = (1.0 - factor) / 3.0;
  factor2 = factor1 - factor;
  buffer->setBetaOrSigmaM(psum->getBetaOrSigmaM() * factor1 - searchPoint[worst]->getBetaOrSigmaM() * factor2);
  buffer->setSigmaE(psum->getSigmaE() * factor1 - searchPoint[worst]->getSigmaE() * factor2);
  buffer->setD1(psum->getD1() * factor1 - searchPoint[worst]->getD1() * factor2);  

  // Prohibits negative values
  if(buffer->getBetaOrSigmaM() < 0.0) buffer->setBetaOrSigmaM(buffer->getBetaOrSigmaM() * (-1.0));
  if(buffer->getSigmaE() < 0.0) buffer->setSigmaE(buffer->getSigmaE() * (-1.0));
  if(buffer->getD1() < 0.0) buffer->setD1(buffer->getD1() * (-1.0));

  buffer->setObjectiveValue(calcLikelihoodMENormal(buffer->getBetaOrSigmaM(),buffer->getSigmaE(),buffer->getD1(),programParameters,data));

  rating = buffer->getObjectiveValue();
  if(buffer->getObjectiveValue() < searchPoint[worst]->getObjectiveValue()){
    cout << "Found a point better than the worst. Replace worst point " << worst + 1 << "." << endl;
    psum->setBetaOrSigmaM(psum->getBetaOrSigmaM() + buffer->getBetaOrSigmaM() - searchPoint[worst]->getBetaOrSigmaM());
    psum->setSigmaE(psum->getSigmaE() + buffer->getSigmaE() - searchPoint[worst]->getSigmaE());
    psum->setD1(psum->getD1() + buffer->getD1() - searchPoint[worst]->getD1());
    copyEstimate(buffer,searchPoint[worst]);
  } else {
    cout << "New point worse than worst point. No replacement." << endl;
  }

  return rating;
}

// scaleMENormalPair:
// Implements reflections of the Downhill Simplex Method for a pair of individuals from the data
// (called by estimateModelParametersNormalMutNormalEnvPair)
double MLOptimizer::scaleMENormalPair(ModelParameters **searchPoint,ModelParameters *buffer,ModelParameters *psum,int worst,double factor,ProgramParameters *programParameters,Data *data,RandomNumberGenerator *rnd,Random *gammarnd,int individual1,int individual2){

  double factor1;
  double factor2;
  double rating;

  factor1 = (1.0 - factor) / 3.0;
  factor2 = factor1 - factor;
  buffer->setBetaOrSigmaM(psum->getBetaOrSigmaM() * factor1 - searchPoint[worst]->getBetaOrSigmaM() * factor2);
  buffer->setSigmaE(psum->getSigmaE() * factor1 - searchPoint[worst]->getSigmaE() * factor2);
  buffer->setD1(psum->getD1() * factor1 - searchPoint[worst]->getD1() * factor2);  

  // Prohibits negative values
  if(buffer->getBetaOrSigmaM() < 0.0) buffer->setBetaOrSigmaM(buffer->getBetaOrSigmaM() * (-1.0));
  if(buffer->getSigmaE() < 0.0) buffer->setSigmaE(buffer->getSigmaE() * (-1.0));
  if(buffer->getD1() < 0.0) buffer->setD1(buffer->getD1() * (-1.0));

  buffer->setObjectiveValue(calcLikelihoodMENormalPair(buffer->getBetaOrSigmaM(),buffer->getSigmaE(),buffer->getD1(),programParameters,data,individual1,individual2));

  rating = buffer->getObjectiveValue();
  if(buffer->getObjectiveValue() < searchPoint[worst]->getObjectiveValue()){
    cout << "Found a point better than the worst. Replace worst point " << worst + 1 << "." << endl;
    psum->setBetaOrSigmaM(psum->getBetaOrSigmaM() + buffer->getBetaOrSigmaM() - searchPoint[worst]->getBetaOrSigmaM());
    psum->setSigmaE(psum->getSigmaE() + buffer->getSigmaE() - searchPoint[worst]->getSigmaE());
    psum->setD1(psum->getD1() + buffer->getD1() - searchPoint[worst]->getD1());
    copyEstimate(buffer,searchPoint[worst]);
  } else {
    cout << "New point worse than worst point. No replacement." << endl;
  }

  return rating;
}

// estimateModelParametersMENormal:
// Implements Downhill Simplex Method to estimate the parameters of the M&E-normal model
void MLOptimizer::estimateModelParametersMENormal(ProgramParameters* programParameters,Data* data,RandomNumberGenerator *rnd,Random *gammarnd){

  cout.precision(programParameters->getOutputPrecision());
  cout << "ESTIMATION BY MAXIMUM LIKELIHOOD" << endl;

  #define TINY 1.0e-10

  double stopConstant = 1.0 - programParameters->getQuality();
  int iteration;

  // Initialize variables
  int i,j,worst,nextWorst,best;
  double stopValue,sum,swap,ysave,ytry;

  estimate->setObjectiveValue(999999999999.0);
  estimate->setQualityValue(0.0);

  for(int repeat = 0; repeat < programParameters->getRepeats(); repeat++){

    iteration = 1;

    // Initialize 4 points in the 3-dimensional space
    cout << "\nRepeat " << repeat + 1 << "/" << programParameters->getRepeats() << endl;
    cout << "\nCreate 4 random points" << endl;
    ModelParameters* buffer = new ModelParameters();
    ModelParameters* psum = new ModelParameters();
    ModelParameters** searchPoint = new ModelParameters*[4];

    psum->setBetaOrSigmaM(0.0);
    psum->setSigmaE(0.0);
    psum->setD1(0.0);
    for(i = 0; i < 4; i++){
      searchPoint[i] = new ModelParameters();
    }
    searchPoint[0]->setBetaOrSigmaM(rnd->getUniformDistributedNumber(programParameters->getSigmaMLo(),programParameters->getSigmaMHi()));
    searchPoint[0]->setSigmaE(rnd->getUniformDistributedNumber(programParameters->getSigmaELo(),programParameters->getSigmaEHi()));
    searchPoint[0]->setD1(rnd->getUniformDistributedNumber(programParameters->getDLo(),programParameters->getDHi()));
    copyEstimate(searchPoint[0],searchPoint[1]);
    copyEstimate(searchPoint[0],searchPoint[2]);
    copyEstimate(searchPoint[0],searchPoint[3]);
    searchPoint[1]->setBetaOrSigmaM(rnd->getUniformDistributedNumber(programParameters->getSigmaMLo(),programParameters->getSigmaMHi()));
    searchPoint[2]->setSigmaE(rnd->getUniformDistributedNumber(programParameters->getSigmaELo(),programParameters->getSigmaEHi()));
    searchPoint[3]->setD1(rnd->getUniformDistributedNumber(programParameters->getDLo(),programParameters->getDHi()));
    for(i = 0; i < 4; i++){
      searchPoint[i]->setObjectiveValue(calcLikelihoodMENormal(searchPoint[i]->getBetaOrSigmaM(),searchPoint[i]->getSigmaE(),searchPoint[i]->getD1(),programParameters,data));
      psum->setBetaOrSigmaM(psum->getBetaOrSigmaM() + searchPoint[i]->getBetaOrSigmaM());
      psum->setSigmaE(psum->getSigmaE() + searchPoint[i]->getSigmaE());
      psum->setD1(psum->getD1() + searchPoint[i]->getD1());
    }

    while(true){

      // Determine the worst, next-worst and best by looping over the points in the simplex  
      best = 0;
      worst = 0;
      nextWorst = 0;
      for(i = 0; i < 4; i++){
        if(searchPoint[i]->getObjectiveValue() <= searchPoint[best]->getObjectiveValue()) best = i;
        if(searchPoint[i]->getObjectiveValue() >= searchPoint[worst]->getObjectiveValue()) worst = i;
      }
      if(worst == 0) nextWorst = 1;
      for(i = 0; i < 4; i++){
        if(searchPoint[i]->getObjectiveValue() >= searchPoint[nextWorst]->getObjectiveValue() && i != worst) nextWorst = i;
      }

      // Output of the current points
      cout << "\nIteration " << iteration << endl;
      cout << "\nPoint:          1       2       3       4" << endl;
      cout << "sigma_m:        " << searchPoint[0]->getBetaOrSigmaM() << "\t" << searchPoint[1]->getBetaOrSigmaM() << "\t" << searchPoint[2]->getBetaOrSigmaM() << "\t" << searchPoint[3]->getBetaOrSigmaM() << endl;
      cout << "sigma_e:        " << searchPoint[0]->getSigmaE() << "\t" << searchPoint[1]->getSigmaE() << "\t" << searchPoint[2]->getSigmaE() << "\t" << searchPoint[3]->getSigmaE() << endl;
      cout << "d:              " << searchPoint[0]->getD1() << "\t" << searchPoint[1]->getD1() << "\t" << searchPoint[2]->getD1() << "\t" << searchPoint[3]->getD1() << endl;
      cout << "-log-like.:     " << searchPoint[0]->getObjectiveValue() << "\t" << searchPoint[1]->getObjectiveValue() << "\t" << searchPoint[2]->getObjectiveValue() << "\t" << searchPoint[3]->getObjectiveValue() << "\t" << endl;
      cout << "\nBest: " << best + 1 << " \tWorst: " << worst + 1 << "\tNext-Worst: " << nextWorst + 1;

      // Stop-criterion
      stopValue = 2.0 * fabs(searchPoint[worst]->getObjectiveValue() - searchPoint[best]->getObjectiveValue()) / (fabs(searchPoint[worst]->getObjectiveValue()) + fabs(searchPoint[best]->getObjectiveValue()) + TINY);
      cout << "\tStop-Dist: " << stopValue - stopConstant << endl;
      if(stopValue < stopConstant){
        cout << "Repeat " << repeat + 1 << ": " << searchPoint[best]->getObjectiveValue() << endl;
        searchPoint[best]->toString(programParameters);
        if(searchPoint[best]->getObjectiveValue() <= estimate->getObjectiveValue()) copyEstimate(searchPoint[best],estimate);          
        break;
      }

      iteration = iteration + 2;

      cout << "\nReflection by factor -1 through the face of the simplex." << endl;
      ytry = scaleMENormal(searchPoint,buffer,psum,worst,-1.0,programParameters,data,rnd,gammarnd);
      if(ytry <= searchPoint[best]->getObjectiveValue()){
        cout << "Point better than best point found. Try an additional reflection by factor 2." << endl;
        ytry = scaleMENormal(searchPoint,buffer,psum,worst,2.0,programParameters,data,rnd,gammarnd);
      } else if (ytry >= searchPoint[nextWorst]->getObjectiveValue()){
        cout << "Reflected point is worse than next-worse. Do a one-dimensional contraction." << endl;
        ysave = searchPoint[worst]->getObjectiveValue();
        ytry = scaleMENormal(searchPoint,buffer,psum,worst,0.5,programParameters,data,rnd,gammarnd);
        if(ytry >= ysave){
          cout << "Can't get rid of that worse point. Contract around best point." << endl;
          for(i = 0; i < 4; i++){
            if(i != best){
              searchPoint[i]->setBetaOrSigmaM(0.5 * (searchPoint[i]->getBetaOrSigmaM() + searchPoint[best]->getBetaOrSigmaM()));
              searchPoint[i]->setSigmaE(0.5 * (searchPoint[i]->getSigmaE() + searchPoint[best]->getSigmaE()));
              searchPoint[i]->setD1(0.5 * (searchPoint[i]->getD1() + searchPoint[best]->getD1()));
              // Prohibits negative values
              if(searchPoint[i]->getBetaOrSigmaM() < 0.0) searchPoint[i]->setBetaOrSigmaM(searchPoint[i]->getBetaOrSigmaM() * (-1.0));
              if(searchPoint[i]->getSigmaE() < 0.0) searchPoint[i]->setSigmaE(searchPoint[i]->getSigmaE() * (-1.0));
              if(searchPoint[i]->getD1() < 0.0) searchPoint[i]->setD1(searchPoint[i]->getD1() * (-1.0));
              searchPoint[i]->setObjectiveValue(calcLikelihoodMENormal(searchPoint[i]->getBetaOrSigmaM(),searchPoint[i]->getSigmaE(),searchPoint[i]->getD1(),programParameters,data));
            }
          }
          // Update psum
          psum->setBetaOrSigmaM(0.0);
          psum->setSigmaE(0.0);
          psum->setD1(0.0);
          for(i = 0; i < 4; i++){
            psum->setBetaOrSigmaM(psum->getBetaOrSigmaM() + searchPoint[i]->getBetaOrSigmaM());
            psum->setSigmaE(psum->getSigmaE() + searchPoint[i]->getSigmaE());
            psum->setD1(psum->getD1() + searchPoint[i]->getD1());
          }
          iteration = iteration + 3;
        }
      } else {
        iteration--;
      }
    }
  }
  estimate->setQualityValue(1.0);
}

// estimateModelParametersMENormalPair:
// Implements Downhill Simplex Method to estimate the parameters of the M&E-normal model for a pair of individuals from the data.
// individual1 and individual2 are indices of the two compared individuals.
void MLOptimizer::estimateModelParametersMENormalPair(ProgramParameters* programParameters,Data* data,RandomNumberGenerator *rnd,Random *gammarnd,int individual1,int individual2){

  cout.precision(programParameters->getOutputPrecision());
  cout << "ESTIMATION BY MAXIMUM LIKELIHOOD FOR PAIR " << individual1 << " - " << individual2 << endl;

  #define TINY 1.0e-10

  double stopConstant = 1.0 - programParameters->getQuality();
  int iteration;

  // Initialize variables
  int i,j,worst,nextWorst,best;
  double stopValue,sum,swap,ysave,ytry;

  estimate->setObjectiveValue(999999999999.0);
  estimate->setQualityValue(0.0);

  for(int repeat = 0; repeat < programParameters->getRepeats(); repeat++){

    iteration = 1;

    // Initialize 4 points in the 3-dimensional space
    cout << "\nRepeat " << repeat + 1 << "/" << programParameters->getRepeats() << endl;
    cout << "\nCreate 4 random points" << endl;
    ModelParameters* buffer = new ModelParameters();
    ModelParameters* psum = new ModelParameters();
    ModelParameters** searchPoint = new ModelParameters*[4];

    psum->setBetaOrSigmaM(0.0);
    psum->setSigmaE(0.0);
    psum->setD1(0.0);
    for(i = 0; i < 4; i++){
      searchPoint[i] = new ModelParameters();
    }
    searchPoint[0]->setBetaOrSigmaM(rnd->getUniformDistributedNumber(programParameters->getSigmaMLo(),programParameters->getSigmaMHi()));
    searchPoint[0]->setSigmaE(rnd->getUniformDistributedNumber(programParameters->getSigmaELo(),programParameters->getSigmaEHi()));
    searchPoint[0]->setD1(rnd->getUniformDistributedNumber(programParameters->getDLo(),programParameters->getDHi()));
    copyEstimate(searchPoint[0],searchPoint[1]);
    copyEstimate(searchPoint[0],searchPoint[2]);
    copyEstimate(searchPoint[0],searchPoint[3]);
    searchPoint[1]->setBetaOrSigmaM(rnd->getUniformDistributedNumber(programParameters->getSigmaMLo(),programParameters->getSigmaMHi()));
    searchPoint[2]->setSigmaE(rnd->getUniformDistributedNumber(programParameters->getSigmaELo(),programParameters->getSigmaEHi()));
    searchPoint[3]->setD1(rnd->getUniformDistributedNumber(programParameters->getDLo(),programParameters->getDHi()));
    for(i = 0; i < 4; i++){
      searchPoint[i]->setObjectiveValue(calcLikelihoodMENormalPair(searchPoint[i]->getBetaOrSigmaM(),searchPoint[i]->getSigmaE(),searchPoint[i]->getD1(),programParameters,data,individual1,individual2));
      psum->setBetaOrSigmaM(psum->getBetaOrSigmaM() + searchPoint[i]->getBetaOrSigmaM());
      psum->setSigmaE(psum->getSigmaE() + searchPoint[i]->getSigmaE());
      psum->setD1(psum->getD1() + searchPoint[i]->getD1());
    }

    while(true){

      // Determine the worst, next-worst and best by looping over the points in the simplex  
      best = 0;
      worst = 0;
      nextWorst = 0;
      for(i = 0; i < 4; i++){
        if(searchPoint[i]->getObjectiveValue() <= searchPoint[best]->getObjectiveValue()) best = i;
        if(searchPoint[i]->getObjectiveValue() >= searchPoint[worst]->getObjectiveValue()) worst = i;
      }
      if(worst == 0) nextWorst = 1;
      for(i = 0; i < 4; i++){
        if(searchPoint[i]->getObjectiveValue() >= searchPoint[nextWorst]->getObjectiveValue() && i != worst) nextWorst = i;
      }

      // Output of the current points
      cout << "\nIteration " << iteration << endl;
      cout << "\nPoint:          1       2       3       4" << endl;
      cout << "sigma_m:        " << searchPoint[0]->getBetaOrSigmaM() << "\t" << searchPoint[1]->getBetaOrSigmaM() << "\t" << searchPoint[2]->getBetaOrSigmaM() << endl;
      cout << "sigma_e:        " << searchPoint[0]->getSigmaE() << "\t" << searchPoint[1]->getSigmaE() << "\t" << searchPoint[2]->getSigmaE() << "\t" << searchPoint[3]->getSigmaE() << endl;
      cout << "d:              " << searchPoint[0]->getD1() << "\t" << searchPoint[1]->getD1() << "\t" << searchPoint[2]->getD1() << "\t" << searchPoint[3]->getD1() << endl;
      cout << "-log-like.:     " << searchPoint[0]->getObjectiveValue() << "\t" << searchPoint[1]->getObjectiveValue() << "\t" << searchPoint[2]->getObjectiveValue() << "\t" << searchPoint[3]->getObjectiveValue() << "\t" << endl;
      cout << "\nBest: " << best + 1 << " \tWorst: " << worst + 1 << "\tNext-Worst: " << nextWorst + 1;
 
      // Stop-criterion
      stopValue = 2.0 * fabs(searchPoint[worst]->getObjectiveValue() - searchPoint[best]->getObjectiveValue()) / (fabs(searchPoint[worst]->getObjectiveValue()) + fabs(searchPoint[best]->getObjectiveValue()) + TINY);
      cout << "\tStop-Dist: " << stopValue - stopConstant << endl;
      if(stopValue < stopConstant){
        cout << "Repeat " << repeat + 1 << ": " << searchPoint[best]->getObjectiveValue() << endl;
        searchPoint[best]->toString(programParameters);
        if(searchPoint[best]->getObjectiveValue() <= estimate->getObjectiveValue()) copyEstimate(searchPoint[best],estimate);          
        break;
      }

      iteration = iteration + 2;

      cout << "\nReflection by factor -1 through the face of the simplex." << endl;
      ytry = scaleMENormalPair(searchPoint,buffer,psum,worst,-1.0,programParameters,data,rnd,gammarnd,individual1,individual2);
      if(ytry <= searchPoint[best]->getObjectiveValue()){
        cout << "Point better than best point found. Try an additional reflection by factor 2." << endl;
        ytry = scaleMENormalPair(searchPoint,buffer,psum,worst,2.0,programParameters,data,rnd,gammarnd,individual1,individual2);
      } else if (ytry >= searchPoint[nextWorst]->getObjectiveValue()){
        cout << "Reflected point is worse than next-worse. Do a one-dimensional contraction." << endl;
        ysave = searchPoint[worst]->getObjectiveValue();
        ytry = scaleMENormalPair(searchPoint,buffer,psum,worst,0.5,programParameters,data,rnd,gammarnd,individual1,individual2);
        if(ytry >= ysave){
          cout << "Can't get rid of that worse point. Contract around best point." << endl;
          for(i = 0; i < 4; i++){
            if(i != best){
              searchPoint[i]->setBetaOrSigmaM(0.5 * (searchPoint[i]->getBetaOrSigmaM() + searchPoint[best]->getBetaOrSigmaM()));
              searchPoint[i]->setSigmaE(0.5 * (searchPoint[i]->getSigmaE() + searchPoint[best]->getSigmaE()));
              searchPoint[i]->setD1(0.5 * (searchPoint[i]->getD1() + searchPoint[best]->getD1()));
              // Prohibits negative values
              if(searchPoint[i]->getBetaOrSigmaM() < 0.0) searchPoint[i]->setBetaOrSigmaM(searchPoint[i]->getBetaOrSigmaM() * (-1.0));
              if(searchPoint[i]->getSigmaE() < 0.0) searchPoint[i]->setSigmaE(searchPoint[i]->getSigmaE() * (-1.0));
              if(searchPoint[i]->getD1() < 0.0) searchPoint[i]->setD1(searchPoint[i]->getD1() * (-1.0));
              searchPoint[i]->setObjectiveValue(calcLikelihoodMENormalPair(searchPoint[i]->getBetaOrSigmaM(),searchPoint[i]->getSigmaE(),searchPoint[i]->getD1(),programParameters,data,individual1,individual2));
            }
          }
          // Update psum
          psum->setBetaOrSigmaM(0.0);
          psum->setSigmaE(0.0);
          psum->setD1(0.0);
          for(i = 0; i < 4; i++){
            psum->setBetaOrSigmaM(psum->getBetaOrSigmaM() + searchPoint[i]->getBetaOrSigmaM());
            psum->setSigmaE(psum->getSigmaE() + searchPoint[i]->getSigmaE());
            psum->setD1(psum->getD1() + searchPoint[i]->getD1());
          }
          iteration = iteration + 3;
        }
      } else {
        iteration--;
      }
    }
  }
  estimate->setQualityValue(1.0);
}

// copyEstimate: Copies a model parameter estimate from the source estimate to the destination estimate
void MLOptimizer::copyEstimate(ModelParameters *source,ModelParameters *destination){
  destination->setAlpha(source->getAlpha());
  destination->setBetaOrSigmaM(source->getBetaOrSigmaM());
  destination->setSigmaE(source->getSigmaE());
  destination->setD1(source->getD1());
  destination->setD2(source->getD2());
  destination->setD3(source->getD3());
  destination->setD4(source->getD4());
  destination->setObjectiveValue(source->getObjectiveValue());
  destination->setQualityValue(source->getQualityValue());

  destination->getResults()->setVarianceWithin1(source->getResults()->getVarianceWithin1());
  destination->getResults()->setVarianceWithin2(source->getResults()->getVarianceWithin2());
  destination->getResults()->setVarianceWithin3(source->getResults()->getVarianceWithin3());
  destination->getResults()->setVariance12(source->getResults()->getVariance12());
  destination->getResults()->setVariance13(source->getResults()->getVariance13());
  destination->getResults()->setVariance23(source->getResults()->getVariance23());
  destination->getResults()->setSkewness12(source->getResults()->getSkewness12());
  destination->getResults()->setSkewness13(source->getResults()->getSkewness13());
  destination->getResults()->setSkewness23(source->getResults()->getSkewness23());
  destination->getResults()->setKurtosis12(source->getResults()->getKurtosis12());
  destination->getResults()->setKurtosis13(source->getResults()->getKurtosis13());
  destination->getResults()->setKurtosis23(source->getResults()->getKurtosis23());
  destination->getResults()->setS1iSkew(source->getResults()->getS1iSkew());
  destination->getResults()->setS2iSkew(source->getResults()->getS2iSkew());
  destination->getResults()->setS1iPearsonsSkew(source->getResults()->getS1iPearsonsSkew());
  destination->getResults()->setS2iPearsonsSkew(source->getResults()->getS2iPearsonsSkew());
  destination->getResults()->setOverallS1iCounter(source->getResults()->getOverallS1iCounter());
  destination->getResults()->setOverallS2iCounter(source->getResults()->getOverallS2iCounter());
  destination->getResults()->setSide(source->getResults()->getSide());
  destination->getResults()->setMirrored(source->getResults()->getMirrored());
}

