/***************************************************************************
 *   Copyright (C) 2006 by Michael Rosskopf                                *
 *   rosskopf@cs.uni-duesseldorf.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*  Description:
A MLOptimizer-object has methods to estimate the parameters of an M&E-normal model from data sets by using a maximum-likelihood method.
*/

#ifndef MLOPTIMIZER_H
#define MLOPTIMIZER_H

#include "programparameters.h"
#include "data.h"
#include "modelparameters.h"
#include "randomnumbergenerator.h"
#include "random.h"
#include "mathfunctions.h"
#include "graphictools.h"

/**
	@author Michael Rosskopf <michael@bioinf.uni-duesseldorf.de>
*/
class MLOptimizer{

  public:

  // Constructors and destructors
  MLOptimizer();
  ~MLOptimizer();

  // getEstimate: Returns the result of the previous parameter estimation
  ModelParameters* getEstimate();

  // factorial: Calculates the factorial
  double factorial(long);

  // poissonPDF: Calculates the density of a Poisson distribution
  double poissonPDF(double,long);

  // normalPDF: Calculates the density of a normal distribution
  double normalPDF(double,double);

  // compoundPoissonMENormalPDF:
  // Calculates the density of a compound Poisson process with normal distributed random variables
  double compoundPoissonMENormalPDF(double,double,double,double,long);

  // calcLikelihoodMENormal:
  // Calculates the likelihood for the parameters of the M&E-normal model given data
  double calcLikelihoodMENormal(double,double,double,ProgramParameters*,Data*);

  // calcLikelihoodMENormalPair:
  // Calculates the likelihood for the parameters of the M&E-normal model given a pair of individuals from the data
  double calcLikelihoodMENormalPair(double,double,double,ProgramParameters*,Data*,int,int);

  // scaleMENormal:
  // Implements reflections of the Downhill Simplex Method
  // (called by estimateModelParametersNormalMutNormalEnv)
  double scaleMENormal(ModelParameters**,ModelParameters*,ModelParameters*,int,double,ProgramParameters*,Data*,RandomNumberGenerator*,Random*);

  // scaleMENormalPair:
  // Implements reflections of the Downhill Simplex Method for a pair of individuals from the data
   // (called by estimateModelParametersNormalMutNormalEnvPair)
  double scaleMENormalPair(ModelParameters**,ModelParameters*,ModelParameters*,int,double,ProgramParameters*,Data*,RandomNumberGenerator*,Random*,int,int);

  // estimateModelParametersMENormal:
  // Implements Downhill Simplex Method to estimate the parameters of the M&E-normal model
  void estimateModelParametersMENormal(ProgramParameters*,Data*,RandomNumberGenerator*,Random*);

  // estimateModelParametersMENormalPair:
  // Implements Downhill Simplex Method to estimate the parameters of the M&E-normal model for a pair of individuals from the data
  void estimateModelParametersMENormalPair(ProgramParameters*,Data*,RandomNumberGenerator*,Random*,int,int);

  // copyEstimate: Copy a model parameter estimate
  void copyEstimate(ModelParameters*,ModelParameters*);

  private:

  ModelParameters* estimate;  // Stores the result of an ml parameter estimation
  double* factorialValue; // Stores the factorials (in order to save computation time)
  MathFunctions* ms; // Used for some mathematical operations
};

#endif
