/***************************************************************************
                          optimizer.cpp  -  description
                             -------------------
    begin                : Thu Apr 28 2005
    copyright            : (C) 2005 by Michael Rosskopf
    email                : rosskopf@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "optimizer.h"

// Constructors and destructors
Optimizer::Optimizer(ProgramParameters *programParameters){
  leftEstimate = new ModelParameters();
  middleEstimate = new ModelParameters();
  rightEstimate = new ModelParameters();
  newEstimate = new ModelParameters();
  estimate = new ModelParameters();
  leftDistribution = new double[programParameters->getNumberOfSimulations()];
  middleDistribution = new double[programParameters->getNumberOfSimulations()];
  rightDistribution = new double[programParameters->getNumberOfSimulations()];
  newDistribution = new double[programParameters->getNumberOfSimulations()];
  distribution = new double[programParameters->getNumberOfSimulations()];
}
Optimizer::~Optimizer(){
  if(leftEstimate != 0) delete leftEstimate;
  if(middleEstimate != 0) delete middleEstimate;
  if(rightEstimate != 0) delete rightEstimate;
  if(newEstimate != 0) delete newEstimate;
  if(estimate != 0) delete estimate;
  if(leftDistribution != 0) delete leftDistribution;
  if(middleDistribution != 0) delete middleDistribution;
  if(rightDistribution != 0) delete rightDistribution;
  if(newDistribution != 0) delete newDistribution;
  if(distribution != 0) delete distribution;
}

// getEstimate: Returns the result of the previous parameter estimation
ModelParameters* Optimizer::getEstimate(){
  return estimate;
}

// setSideAndMirrored: Only used for the M-gamma model!
//                     All estimate buffer used by optimizer are setted to the used distribution,
//                     specified in targetValues.
//                     (normal gamma distribution ore mirrored one)
 void Optimizer::setSideAndMirrored(DistributionMeasures *targetValues){
  if(targetValues->getSide() == 0){
    leftEstimate->getResults()->setSide(0);
    middleEstimate->getResults()->setSide(0);
    rightEstimate->getResults()->setSide(0);
    newEstimate->getResults()->setSide(0);
    estimate->getResults()->setSide(0);
  } else {
    leftEstimate->getResults()->setSide(1);
    middleEstimate->getResults()->setSide(1);
    rightEstimate->getResults()->setSide(1);
    newEstimate->getResults()->setSide(1);
    estimate->getResults()->setSide(1);
  }
  if(targetValues->getMirrored() == 0){
    // Normal gamma distribution
    leftEstimate->getResults()->setMirrored(0);
    middleEstimate->getResults()->setMirrored(0);
    rightEstimate->getResults()->setMirrored(0);
    newEstimate->getResults()->setMirrored(0);
    estimate->getResults()->setMirrored(0);
  } else {
    // Mirrored gamma distribution
    leftEstimate->getResults()->setMirrored(1);
    middleEstimate->getResults()->setMirrored(1);
    rightEstimate->getResults()->setMirrored(1);
    newEstimate->getResults()->setMirrored(1);
    estimate->getResults()->setMirrored(1);
  }
}

// estimateMGammaParameters: Estimates the parameters of the M-gamma model for data
void Optimizer::estimateMGammaParameters(ProgramParameters *programParameters,Data *data,ThreeTree *threeTree,DistributionMeasures *targetValues,RandomNumberGenerator *rnd,Random *gammarnd){

  bool stop = false;
  bool limitFound = false;
  double alphaLIS = 0.0;
  double alphaRIS = 0.0;
  int mirroredLIS;
  int mirroredRIS;

  // Initialize the quality with 1 (value is changed to 0, if limit is exceeded)
  estimate->setNull();
  estimate->setQualityValue(1.0);

  for(int i = 0; i < programParameters->getNumberOfSets(); i++){

    if(!stop){

      if(programParameters->getNumberOfSets() == 2){
        if(i == 0){
          targetValues->setSide(0);
          if(programParameters->getSkewType() == 1){
            if(targetValues->getS1iSkew() < 0.0){
              targetValues->setMirrored(0);
            } else {
              targetValues->setMirrored(1);
            }
          }
          if(programParameters->getSkewType() == 2){
            if(targetValues->getS1iPearsonsSkew() < 0.0){
              targetValues->setMirrored(0);
            } else {
              targetValues->setMirrored(1);
            }
          }
          cout << "\nUSE SAMPLE-1-INTERMEDIATE GENES" << endl;
        }
        if(i == 1){
          targetValues->setSide(1);
          if(programParameters->getSkewType() == 1){
            if(targetValues->getS2iSkew() > 0.0){
              targetValues->setMirrored(0);
            } else {
              targetValues->setMirrored(1);
            }
          }
          if(programParameters->getSkewType() == 2){
            if(targetValues->getS2iPearsonsSkew() > 0.0){
              targetValues->setMirrored(0);
            } else {
              targetValues->setMirrored(1);
            }
          }
          cout << "\nUSE SAMPLE-2-INTERMEDIATE GENES" << endl;
        }
      }
      setSideAndMirrored(targetValues); // Is it necessary to mirror the gamma distribution

      // Init estimates, so that the left and the right ones have different signs
      middleEstimate->setQualityValue(0.0);

      if(programParameters->getOutputMode() > 1) cout << "\nCalculating a root bracketing" << endl;
      if(programParameters->getOutputMode() > 1) cout << "\nLeft estimate:" << endl;
      leftEstimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getStandardLowerBoundAlpha());
      threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
      mGammaObjectiveFunction(programParameters,leftEstimate,targetValues);

      if(programParameters->getOutputMode() > 1) leftEstimate->toString(programParameters);
      if(programParameters->getOutputMode() > 1) cout << "\nRight estimate:" << endl;
      rightEstimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getStandardUpperBoundAlpha());
      threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
      mGammaObjectiveFunction(programParameters,rightEstimate,targetValues);

      if(programParameters->getOutputMode() > 1) rightEstimate->toString(programParameters);

      // Root between the left and right estimate? Stop trying, if alpha outside limits.
      limitFound = false;
      while(!limitFound && !stop){
        if(targetValues->getSide() == 0){
          if(targetValues->getMirrored() == 0){
            if(leftEstimate->getObjectiveValue() >= 0.0){
              leftEstimate->setAlpha(leftEstimate->getAlpha() / 2.0);
              if(leftEstimate->getAlpha() < programParameters->getMinimalAlpha()) leftEstimate->setAlpha(programParameters->getMinimalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,leftEstimate,targetValues);
              if(leftEstimate->getAlpha() == programParameters->getMinimalAlpha() && leftEstimate->getObjectiveValue() >= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMinimalAlpha());
              }
            }
            if(rightEstimate->getObjectiveValue() <= 0.0){
              rightEstimate->setAlpha(rightEstimate->getAlpha() * 2.0);
              if(rightEstimate->getAlpha() > programParameters->getMaximalAlpha()) rightEstimate->setAlpha(programParameters->getMaximalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,rightEstimate,targetValues);
              if(rightEstimate->getAlpha() == programParameters->getMaximalAlpha() && rightEstimate->getObjectiveValue() <= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMaximalAlpha());
              }
            }
          } else {
            if(leftEstimate->getObjectiveValue() <= 0.0){
              leftEstimate->setAlpha(leftEstimate->getAlpha() / 2.0);
              if(leftEstimate->getAlpha() < programParameters->getMinimalAlpha()) leftEstimate->setAlpha(programParameters->getMinimalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,leftEstimate,targetValues);
              if(leftEstimate->getAlpha() == programParameters->getMinimalAlpha() && leftEstimate->getObjectiveValue() <= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMinimalAlpha());
              }
            }
            if(rightEstimate->getObjectiveValue() >= 0.0){
              rightEstimate->setAlpha(rightEstimate->getAlpha() * 2.0);
              if(rightEstimate->getAlpha() > programParameters->getMaximalAlpha()) rightEstimate->setAlpha(programParameters->getMaximalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,rightEstimate,targetValues);
              if(rightEstimate->getAlpha() == programParameters->getMaximalAlpha() && rightEstimate->getObjectiveValue() >= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMaximalAlpha());
              }
            }
          }
        }
        if(targetValues->getSide() == 1){
          if(targetValues->getMirrored() == 0){
             if(leftEstimate->getObjectiveValue() <= 0.0){
              leftEstimate->setAlpha(leftEstimate->getAlpha() / 2.0);
              if(leftEstimate->getAlpha() < programParameters->getMinimalAlpha()) leftEstimate->setAlpha(programParameters->getMinimalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,leftEstimate,targetValues);
              if(leftEstimate->getAlpha() == programParameters->getMinimalAlpha() && leftEstimate->getObjectiveValue() <= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMinimalAlpha());
              }
            }
            if(rightEstimate->getObjectiveValue() >= 0.0){
              rightEstimate->setAlpha(rightEstimate->getAlpha() * 2.0);
              if(rightEstimate->getAlpha() > programParameters->getMaximalAlpha()) rightEstimate->setAlpha(programParameters->getMaximalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,rightEstimate,targetValues);
              if(rightEstimate->getAlpha() == programParameters->getMaximalAlpha() && rightEstimate->getObjectiveValue() >= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMaximalAlpha());
              }
            }
          } else {
            if(leftEstimate->getObjectiveValue() >= 0.0){
              leftEstimate->setAlpha(leftEstimate->getAlpha() / 2.0);
              if(leftEstimate->getAlpha() < programParameters->getMinimalAlpha()) leftEstimate->setAlpha(programParameters->getMinimalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,leftEstimate,targetValues);
              if(leftEstimate->getAlpha() == programParameters->getMinimalAlpha() && leftEstimate->getObjectiveValue() >= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMinimalAlpha());
              }
            }
            if(rightEstimate->getObjectiveValue() <= 0.0){
              rightEstimate->setAlpha(rightEstimate->getAlpha() * 2.0);
              if(rightEstimate->getAlpha() > programParameters->getMaximalAlpha()) rightEstimate->setAlpha(programParameters->getMaximalAlpha());
              threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
              mGammaObjectiveFunction(programParameters,rightEstimate,targetValues);
              if(rightEstimate->getAlpha() == programParameters->getMaximalAlpha() && rightEstimate->getObjectiveValue() <= 0.0){
                stop = true;
                estimate->setParametersMGammaGivenAlpha(programParameters,targetValues,programParameters->getMaximalAlpha());
              }
            }
          }
        }
        if((leftEstimate->getObjectiveValue() < 0.0 && rightEstimate->getObjectiveValue() > 0.0) || (leftEstimate->getObjectiveValue() > 0.0 && rightEstimate->getObjectiveValue() < 0.0)){
          limitFound = true;
        }
        cout << "\nBracketing:" << endl;
        cout << "Alpha's:    " << leftEstimate->getAlpha() << " <-L R-> " << rightEstimate->getAlpha() << endl;
        cout << "Distances: (" << leftEstimate->getObjectiveValue() << " ; " << rightEstimate->getObjectiveValue() << ")" << endl;
      }

      if(!stop){
        cout << "\nBracketing has been found. Starting root search." << endl;
        // Bracketing
        while((rightEstimate->getAlpha() - leftEstimate->getAlpha()) > (1.0 - programParameters->getQuality())){
          double newAlpha = rnd->getUniformDistributedNumber(leftEstimate->getAlpha(),rightEstimate->getAlpha());           // By random
          if(programParameters->getOutputMode() > 1) cout << "\nNew middle estimate for alpha: " << newAlpha << endl;
          middleEstimate->setParametersMGammaGivenAlpha(programParameters,targetValues,newAlpha);
          threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations() / 10.0,middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
          mGammaObjectiveFunction(programParameters,middleEstimate,targetValues);
          if(programParameters->getOutputMode() > 1) middleEstimate->toString(programParameters);
          if((middleEstimate->getObjectiveValue() < 0.0 && leftEstimate->getObjectiveValue() < 0.0) || (middleEstimate->getObjectiveValue() > 0.0 && leftEstimate->getObjectiveValue() > 0.0)){
            copyEstimate(middleEstimate,leftEstimate);
          } else {
            copyEstimate(middleEstimate,rightEstimate);
          }
        }

        // Fine tuning
        if(programParameters->getOutputMode() > 1) cout << "\nFine tuning" << endl;
        middleEstimate->setQualityValue(0.0);

        if(programParameters->getOutputMode() > 1) cout << "\nLeft estimate:" << endl;
        leftEstimate->setParametersMGammaGivenAlpha(programParameters,targetValues,middleEstimate->getAlpha() * 0.9);
        if(leftEstimate->getAlpha() < programParameters->getMinimalAlpha()) leftEstimate->setAlpha(programParameters->getMinimalAlpha());
        threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
        mGammaObjectiveFunction(programParameters,leftEstimate,targetValues);
        if(programParameters->getOutputMode() > 1) leftEstimate->toString(programParameters);

        if(programParameters->getOutputMode() > 1) cout << "\nRight estimate:" << endl;
        rightEstimate->setParametersMGammaGivenAlpha(programParameters,targetValues,middleEstimate->getAlpha() * 1.1);
        if(rightEstimate->getAlpha() > programParameters->getMaximalAlpha()) rightEstimate->setAlpha(programParameters->getMaximalAlpha());
        threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
        mGammaObjectiveFunction(programParameters,rightEstimate,targetValues);
        if(programParameters->getOutputMode() > 1) rightEstimate->toString(programParameters);

        while((rightEstimate->getAlpha() - leftEstimate->getAlpha()) > (1.0 - programParameters->getQuality())){
          double newAlpha = rnd->getUniformDistributedNumber(leftEstimate->getAlpha(),rightEstimate->getAlpha());           // By random
          if(programParameters->getOutputMode() > 1) cout << "\nNew middle estimate for alpha: " << newAlpha << endl;
          middleEstimate->setParametersMGammaGivenAlpha(programParameters,targetValues,newAlpha);
          threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
          mGammaObjectiveFunction(programParameters,middleEstimate,targetValues);
          if(programParameters->getOutputMode() > 1) middleEstimate->toString(programParameters);
          if((middleEstimate->getObjectiveValue() < 0.0 && leftEstimate->getObjectiveValue() < 0.0) || (middleEstimate->getObjectiveValue() > 0.0 && leftEstimate->getObjectiveValue() > 0.0)){
            copyEstimate(middleEstimate,leftEstimate);
          } else {
            copyEstimate(middleEstimate,rightEstimate);
          }
        }
        copyEstimate(middleEstimate,estimate);
      }

      if(i == 0){
        alphaLIS = estimate->getAlpha();
        mirroredLIS = estimate->getResults()->getMirrored();
      }
      if(i == 1){
        alphaRIS = estimate->getAlpha();
        mirroredRIS = estimate->getResults()->getMirrored();
      }
    }
  }

  if(programParameters->getNumberOfSets() == 2){

    // Mean of both alpha-estimates
    // Weight mean over the size of LIS and RIS subsets
    if(!stop) estimate->setAlpha(alphaLIS * (double(targetValues->getOverallS1iCounter()) / (double(targetValues->getOverallS1iCounter()) + double(targetValues->getOverallS2iCounter()))) + alphaRIS * (double(targetValues->getOverallS2iCounter()) / (double(targetValues->getOverallS1iCounter()) + double(targetValues->getOverallS2iCounter()))));

    // Write the two alpha estimators to a file
//    ofstream file("_both_alpha-estimates.txt",ios::app);
//    file << alphaLIS << "\t" << alphaRIS << endl;
//    file.close();

    if(!stop) threeTree->simulationMGamma(programParameters,data,programParameters->getNumberOfSimulations(),estimate,estimate->getResults(),rnd,gammarnd);
    // Set side in estimate to 2 if two sides have been used
    estimate->getResults()->setSide(2);
    // Set mirrored in best estimate to 2, if a normal and a mirrored distribution have been used
    if(mirroredLIS != mirroredRIS) estimate->getResults()->setMirrored(2);
  }

  // Quality
  if(stop){
    estimate->setQualityValue(0.0);
  } else {
    estimate->setQualityValue(1.0);
  }
}

// mGammaObjectiveFunction: Objective function of the optimisation method of the M-gamma model
void Optimizer::mGammaObjectiveFunction(ProgramParameters *programParameters,ModelParameters *modelParameters,DistributionMeasures *target){

  // Initialization
  double selectionValue = 0;
  double buffer = 0;
  double buffer2 = 0;

  // Root find method - Optimization on S1iSkew
  if(target->getSide() == 0){
    if(programParameters->getSkewType() == 1) buffer = modelParameters->getResults()->getS1iSkew() - target->getS1iSkew();
    if(programParameters->getSkewType() == 2) buffer = modelParameters->getResults()->getS1iPearsonsSkew() - target->getS1iPearsonsSkew();
    modelParameters->setObjectiveValue(buffer);
    if(buffer < 0) buffer = buffer * (-1);
    modelParameters->setQualityValue(1.0 / (1.0 + buffer));
  }
  // Root find method - Optimization on S2iSkew
  if(target->getSide() == 1){
    if(programParameters->getSkewType() == 1) buffer = modelParameters->getResults()->getS2iSkew() - target->getS2iSkew();
    if(programParameters->getSkewType() == 2) buffer = modelParameters->getResults()->getS2iPearsonsSkew() - target->getS2iPearsonsSkew();
    modelParameters->setObjectiveValue(buffer);
    if(buffer < 0) buffer = buffer * (-1);
    modelParameters->setQualityValue(1.0 / (1.0 + buffer));
  }
}

// estimateMENormalParameters: Estimates the parameters of the M&E-normal model for data
void Optimizer::estimateMENormalParameters(ProgramParameters *programParameters,Data *data,ThreeTree *threeTree,double targetHistogramPlot[],double lowerBound,double upperBound,DistributionMeasures *targetValues,RandomNumberGenerator *rnd,Random *gammarnd){

  // Initialization
  int intervalShiftMode = 0; // Mode for shifting the interval to find the Brent initial values
  double multiplicator;
  bool stop = false;
  do{
    multiplicator = 1.0 + rnd->getNormalDistributedNumber(0.0,0.1);
  } while(multiplicator < 0.5 || multiplicator > 1.5);
  double newError;
  double maximalSigmaE = pow(0.5 * targetValues->getVariance12(),0.5) - (1.0 - programParameters->getQuality());
  leftEstimate->setQualityValue(0.0);
  middleEstimate->setQualityValue(0.0);
  rightEstimate->setQualityValue(0.0);
  newEstimate->setQualityValue(0.0);
  estimate->setQualityValue(0.0);
  estimate->setObjectiveValue(9999999999.0);

  // Init estimates, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)
  if(programParameters->getOutputMode() > 1) cout << "\nCalculating startvalues, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)" << endl;

  leftEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.0);
  if((leftEstimate->getD1() + leftEstimate->getD2() + leftEstimate->getD3() + leftEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationME(programParameters,data,leftDistribution,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,leftEstimate,leftDistribution,targetHistogramPlot,lowerBound,upperBound);
  middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.5 * maximalSigmaE);
  if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationME(programParameters,data,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
  rightEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,maximalSigmaE);
  if((rightEstimate->getD1() + rightEstimate->getD2() + rightEstimate->getD3() + rightEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationME(programParameters,data,rightDistribution,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,rightEstimate,rightDistribution,targetHistogramPlot,lowerBound,upperBound);
  if(programParameters->getOutputMode() > 1){
    cout << "\nLeft estimate:   " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
    cout << "Middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
    cout << "Right estimate:   " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
  }

  // Shift the interval to find a minimum
  while(!stop && !(leftEstimate->getObjectiveValue() > middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()) && ((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality()))){

    // left < middle < right
    if(leftEstimate->getObjectiveValue() <= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() <= rightEstimate->getObjectiveValue()) intervalShiftMode = 1;
    // left > middle > right
    if(leftEstimate->getObjectiveValue() >= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() >= rightEstimate->getObjectiveValue()) intervalShiftMode = 2;
    // middle > (left && right)
    if(leftEstimate->getObjectiveValue() < middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() > rightEstimate->getObjectiveValue()) intervalShiftMode = 3;

    // left <= middle <= right
    if(intervalShiftMode == 1){
      cout << "\nLeft estimate <= Middle estimate <= Right estimate" << endl;
      copyEstimate(middleEstimate,rightEstimate);
      middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
    }
    // left >= middle >= right
    if(intervalShiftMode == 2){
      cout << "\nLeft estimate >= Middle estimate >= Right estimate" << endl;
      copyEstimate(middleEstimate,leftEstimate);
      middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
    }
    // middle > (left && right)
    if(intervalShiftMode == 3){
      cout << "\nMiddle estimate > (Left estimate && Right estimate)" << endl;
      if(leftEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()){
        // left < right
        copyEstimate(middleEstimate,rightEstimate);
        middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
      } else {
        // left > right
        copyEstimate(middleEstimate,leftEstimate);
        middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
      }
    }
    // Simulation for new middle estimate
    if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
      threeTree->simulationME(programParameters,data,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
    } else {
      cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
      stop = true;
    }
    mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
    if(programParameters->getOutputMode() > 1){
      cout << "\nNew left estimate:     " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
      cout << "New middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
      cout << "New right estimate:    " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
    }
  }

  // Brents method for finding the minimum
  if(!stop){
    if(programParameters->getOutputMode() > 1) cout << "\nInitial values found. Start Brents Method." << endl;
    int i = 0;
    while((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality())){
      i++;
      if(rnd->getUniformDistributedNumber(0.0,1.0) < 0.5){
        newError = middleEstimate->getSigmaE() - 0.38197 * (middleEstimate->getSigmaE() - leftEstimate->getSigmaE());
        newEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationME(programParameters,data,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,leftEstimate);
        } else {
          copyEstimate(middleEstimate,rightEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      } else {
        newError = middleEstimate->getSigmaE() + 0.38197 * (rightEstimate->getSigmaE() - middleEstimate->getSigmaE());
        newEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationME(programParameters,data,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,rightEstimate);
        } else {
          copyEstimate(middleEstimate,leftEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      }
      if(programParameters->getOutputMode() > 1){
        cout << "\n----------------------------------------------------------------" << endl;
        cout << "\nIteration " << i;
        cout << "\nNew sigma_e estimate:  " << newEstimate->getSigmaE() << " \tDistance: " << newEstimate->getObjectiveValue() << endl;
        cout << "\nNew left estimate:   " << leftEstimate->getSigmaE() << " \tDistance: " << leftEstimate->getObjectiveValue() << endl;
        cout << "New middle estimate: " << middleEstimate->getSigmaE() << " \tDistance: " << middleEstimate->getObjectiveValue() << endl;
        cout << "New right estimate:  " << rightEstimate->getSigmaE() << " \tDistance: " << rightEstimate->getObjectiveValue() << endl;
      }
    }
    // Determining best estimate and set its quality equal 1.
    if(estimate->getObjectiveValue() > leftEstimate->getObjectiveValue()){
      copyEstimate(leftEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > middleEstimate->getObjectiveValue()){
      copyEstimate(middleEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > rightEstimate->getObjectiveValue()){
      copyEstimate(rightEstimate,estimate);
    }
  }

  if(!stop){
    estimate->setQualityValue(1.0);
  } else {
    estimate->setQualityValue(0.0);
  }
}

// estimateMENormalParametersForPairs: Estimates the parameters of the M&E-normal model for a pair of individuals specified by their gene expression difference distribution targetHistogramPlot
void Optimizer::estimateMENormalParametersForPairs(ProgramParameters *programParameters,ThreeTree *threeTree,double targetHistogramPlot[],double lowerBound,double upperBound,DistributionMeasures *targetValues,RandomNumberGenerator *rnd,Random *gammarnd){

  // Initialization
  int intervalShiftMode = 0; // Mode for shifting the interval to find the Brent initial values
  double multiplicator;
  bool stop = false;
  do{
    multiplicator = 1.0 + rnd->getNormalDistributedNumber(0.0,0.1);
  } while(multiplicator < 0.5 || multiplicator > 1.5);
  double newError;
  double maximalSigmaE = pow(0.5 * targetValues->getVariance12(),0.5) - (1.0 - programParameters->getQuality());
  leftEstimate->setQualityValue(0.0);
  middleEstimate->setQualityValue(0.0);
  rightEstimate->setQualityValue(0.0);
  newEstimate->setQualityValue(0.0);
  estimate->setQualityValue(0.0);
  estimate->setObjectiveValue(9999999999.0);

  // Init estimates, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)
  if(programParameters->getOutputMode() > 1) cout << "\nCalculating startvalues, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)" << endl;

  leftEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.0); 
  if((leftEstimate->getD1() + leftEstimate->getD2() + leftEstimate->getD3() + leftEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationMEForPairs(programParameters,leftDistribution,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,leftEstimate,leftDistribution,targetHistogramPlot,lowerBound,upperBound);
  middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.5 * maximalSigmaE);
  if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationMEForPairs(programParameters,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
  rightEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,maximalSigmaE);
  if((rightEstimate->getD1() + rightEstimate->getD2() + rightEstimate->getD3() + rightEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationMEForPairs(programParameters,rightDistribution,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,rightEstimate,rightDistribution,targetHistogramPlot,lowerBound,upperBound);
  if(programParameters->getOutputMode() > 1){
    cout << "\nLeft estimate:   " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
    cout << "Middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
    cout << "Right estimate:   " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
  }

  // Shift the interval to find a minimum
  while(!stop && !(leftEstimate->getObjectiveValue() > middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()) && ((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality()))){

    // left < middle < right
    if(leftEstimate->getObjectiveValue() <= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() <= rightEstimate->getObjectiveValue()) intervalShiftMode = 1;
    // left > middle > right
    if(leftEstimate->getObjectiveValue() >= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() >= rightEstimate->getObjectiveValue()) intervalShiftMode = 2;
    // middle > (left && right)
    if(leftEstimate->getObjectiveValue() < middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() > rightEstimate->getObjectiveValue()) intervalShiftMode = 3;

    // left <= middle <= right
    if(intervalShiftMode == 1){
      cout << "\nLeft estimate <= Middle estimate <= Right estimate" << endl;
      copyEstimate(middleEstimate,rightEstimate);
      middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
    }
    // left >= middle >= right
    if(intervalShiftMode == 2){
      cout << "\nLeft estimate >= Middle estimate >= Right estimate" << endl;
      copyEstimate(middleEstimate,leftEstimate);
      middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
    }
    // middle > (left && right)
    if(intervalShiftMode == 3){
      cout << "\nMiddle estimate > (Left estimate && Right estimate)" << endl;
      if(leftEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()){
        // left < right
        copyEstimate(middleEstimate,rightEstimate);
        middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
      } else {
        // left > right
        copyEstimate(middleEstimate,leftEstimate);
        middleEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
      }
    }
    // Simulation for new middle estimate
    if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
      threeTree->simulationMEForPairs(programParameters,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
    } else {
      cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
      stop = true;
    }
    mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
    if(programParameters->getOutputMode() > 1){
      cout << "\nNew left estimate:   " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
      cout << "New middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
      cout << "New right estimate:   " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
    }
  }

  // Brents method for finding the minimum
  if(!stop){
    if(programParameters->getOutputMode() > 1) cout << "\nInitial values found. Start Brents Method." << endl;
    int i = 0;
    while((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality())){
      i++;
      if(rnd->getUniformDistributedNumber(0.0,1.0) < 0.5){
        newError = middleEstimate->getSigmaE() - 0.38197 * (middleEstimate->getSigmaE() - leftEstimate->getSigmaE());
        newEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationMEForPairs(programParameters,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,leftEstimate);
        } else {
          copyEstimate(middleEstimate,rightEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      } else {
        newError = middleEstimate->getSigmaE() + 0.38197 * (rightEstimate->getSigmaE() - middleEstimate->getSigmaE());
        newEstimate->setParametersMENormalGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationMEForPairs(programParameters,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,rightEstimate);
        } else {
          copyEstimate(middleEstimate,leftEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      }
      if(programParameters->getOutputMode() > 1){
        cout << "\n----------------------------------------------------------------" << endl;
        cout << "\nIteration " << i;
        cout << "\nNew sigma_e estimate:  " << newEstimate->getSigmaE() << " \tDistance: " << newEstimate->getObjectiveValue() << endl;
        cout << "\nNew left estimate:   " << leftEstimate->getSigmaE() << " \tDistance: " << leftEstimate->getObjectiveValue() << endl;
        cout << "New middle estimate: " << middleEstimate->getSigmaE() << " \tDistance: " << middleEstimate->getObjectiveValue() << endl;
        cout << "New right estimate:  " << rightEstimate->getSigmaE() << " \tDistance: " << rightEstimate->getObjectiveValue() << endl;
      }
    }
    // Determining best estimate and set its quality equal 1.
    if(estimate->getObjectiveValue() > leftEstimate->getObjectiveValue()){
      copyEstimate(leftEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > middleEstimate->getObjectiveValue()){
      copyEstimate(middleEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > rightEstimate->getObjectiveValue()){
      copyEstimate(rightEstimate,estimate);
    }
  }

  if(!stop){
    estimate->setQualityValue(1.0);
  } else {
    estimate->setQualityValue(0.0);
  }
}

// estimateMEExtremeParameters: Estimates the parameters of the M&E-extreme model for data
void Optimizer::estimateMEExtremeParameters(ProgramParameters *programParameters,Data *data,ThreeTree *threeTree,double targetHistogramPlot[],double lowerBound,double upperBound,DistributionMeasures *targetValues,RandomNumberGenerator *rnd,Random *gammarnd){

  // Initialization
  int intervalShiftMode = 0; // Mode for shifting the interval to find the Brent initial values
  bool stop = false;
  double newError;
  double maximalSigmaE = newEstimate->calcMEExtremeMaximalSigmaE(programParameters,targetValues) * programParameters->getQuality();

  leftEstimate->setQualityValue(0.0);
  middleEstimate->setQualityValue(0.0);
  rightEstimate->setQualityValue(0.0);
  newEstimate->setQualityValue(0.0);
  estimate->setQualityValue(0.0);
  estimate->setObjectiveValue(9999999999.0);

  // Init estimates, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)
  if(programParameters->getOutputMode() > 1) cout << "\nCalculating startvalues, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)" << endl;
  leftEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.0);
  if((leftEstimate->getD1() + leftEstimate->getD2() + leftEstimate->getD3() + leftEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationME(programParameters,data,leftDistribution,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 = " << (leftEstimate->getD1() + leftEstimate->getD2() + leftEstimate->getD3() + leftEstimate->getD4()) << " > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,leftEstimate,leftDistribution,targetHistogramPlot,lowerBound,upperBound);
  cout << "\nLeft Estimate:" << endl;
  leftEstimate->toString(programParameters);	

  middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.5 * maximalSigmaE);
  if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationME(programParameters,data,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4  = " << (middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) << " > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
  cout << "\nMiddle Estimate:" << endl;
  middleEstimate->toString(programParameters);	

  rightEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,maximalSigmaE);
  if((rightEstimate->getD1() + rightEstimate->getD2() + rightEstimate->getD3() + rightEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationME(programParameters,data,rightDistribution,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4  = " << (rightEstimate->getD1() + rightEstimate->getD2() + rightEstimate->getD3() + rightEstimate->getD4()) << " > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,rightEstimate,rightDistribution,targetHistogramPlot,lowerBound,upperBound);
  cout << "\nRight Estimate:" << endl;
  rightEstimate->toString(programParameters);	

  if(programParameters->getOutputMode() > 1){
    cout << "\nLeft estimate:   " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
    cout << "Middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
    cout << "Right estimate:   " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
  }

  // Shift the interval to find a minimum
  while(!stop && !(leftEstimate->getObjectiveValue() > middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()) && ((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality()))){

    // left < middle < right
    if(leftEstimate->getObjectiveValue() <= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() <= rightEstimate->getObjectiveValue()) intervalShiftMode = 1;
    // left > middle > right
    if(leftEstimate->getObjectiveValue() >= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() >= rightEstimate->getObjectiveValue()) intervalShiftMode = 2;
    // middle > (left && right)
    if(leftEstimate->getObjectiveValue() < middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() > rightEstimate->getObjectiveValue()) intervalShiftMode = 3;

    // left <= middle <= right
    if(intervalShiftMode == 1){
      cout << "\nLeft estimate <= Middle estimate <= Right estimate" << endl;
      copyEstimate(middleEstimate,rightEstimate);
      middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
    }
    // left >= middle >= right
    if(intervalShiftMode == 2){
      cout << "\nLeft estimate >= Middle estimate >= Right estimate" << endl;
      copyEstimate(middleEstimate,leftEstimate);
      middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
    }
    // middle > (left && right)
    if(intervalShiftMode == 3){
      cout << "\nMiddle estimate > (Left estimate && Right estimate)" << endl;
      if(leftEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()){
        // left < right
        copyEstimate(middleEstimate,rightEstimate);
        middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
      } else {
        // left > right
        copyEstimate(middleEstimate,leftEstimate);
        middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
      }
    }
    // Simulation for new middle estimate
    if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
      threeTree->simulationME(programParameters,data,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
    } else {
      cout << "d_1 + d_2 + d_3 + d_4  > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
      stop = true;
    }
    mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
    if(programParameters->getOutputMode() > 1){
      cout << "\nNew left estimate:   " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
      cout << "New middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
      cout << "New right estimate:   " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
    }
  }

  // Brents method for finding the minimum
  if(!stop){
    if(programParameters->getOutputMode() > 1) cout << "\nInitial values found. Start Brents Method." << endl;
    int i = 0;
    while((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality())){
      i++;
      if(rnd->getUniformDistributedNumber(0.0,1.0) < 0.5){
        newError = middleEstimate->getSigmaE() - 0.38197 * (middleEstimate->getSigmaE() - leftEstimate->getSigmaE());
        newEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationME(programParameters,data,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
  cout << "\nNew Estimate:" << endl;
  newEstimate->toString(programParameters);	
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,leftEstimate);
        } else {
          copyEstimate(middleEstimate,rightEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      } else {
        newError = middleEstimate->getSigmaE() + 0.38197 * (rightEstimate->getSigmaE() - middleEstimate->getSigmaE());
        newEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationME(programParameters,data,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
  cout << "\nNew Estimate:" << endl;
  newEstimate->toString(programParameters);	
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,rightEstimate);
        } else {
          copyEstimate(middleEstimate,leftEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      }
      if(programParameters->getOutputMode() > 1){
        cout << "\n----------------------------------------------------------------" << endl;
        cout << "\nIteration " << i;
        cout << "\nNew sigma_e estimate:  " << newEstimate->getSigmaE() << " \tDistance: " << newEstimate->getObjectiveValue() << endl;
        cout << "\nNew left estimate:   " << leftEstimate->getSigmaE() << " \tDistance: " << leftEstimate->getObjectiveValue() << endl;
        cout << "New middle estimate: " << middleEstimate->getSigmaE() << " \tDistance: " << middleEstimate->getObjectiveValue() << endl;
        cout << "New right estimate:  " << rightEstimate->getSigmaE() << " \tDistance: " << rightEstimate->getObjectiveValue() << endl;
      }
    }
    // Determining best estimate and set its quality equal 1.
    if(estimate->getObjectiveValue() > leftEstimate->getObjectiveValue()){
      copyEstimate(leftEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > middleEstimate->getObjectiveValue()){
      copyEstimate(middleEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > rightEstimate->getObjectiveValue()){
      copyEstimate(rightEstimate,estimate);
    }
  }

  // Check for nan. Model can not fit the data in this case.
  if(isnan(estimate->getBetaOrSigmaM())){
    stop = true;
  }
  if(isnan(estimate->getD1())){
    stop = true;
  }
  if(isnan(estimate->getD2())){
    stop = true;
  }

  if(!stop){
    estimate->setQualityValue(1.0);
  } else {
    cout << "\nModel can not fit the data!" << endl;
    estimate->setQualityValue(0.0);
  }
}

// estimateMEExtremeParametersForPairs: Estimates the parameters of the M&E-extreme model for a pair of individuals specified by their gene expression difference distribution targetHistogramPlot
void Optimizer::estimateMEExtremeParametersForPairs(ProgramParameters *programParameters,ThreeTree *threeTree,double targetHistogramPlot[],double lowerBound,double upperBound,DistributionMeasures *targetValues,RandomNumberGenerator *rnd,Random *gammarnd){

  // Initialization
  int intervalShiftMode = 0; // Mode for shifting the interval to find the Brent initial values
  bool stop = false;
  double newError;
  double maximalSigmaE = newEstimate->calcMEExtremeMaximalSigmaE(programParameters,targetValues) - (1.0 - programParameters->getQuality());
  leftEstimate->setQualityValue(0.0);
  middleEstimate->setQualityValue(0.0);
  rightEstimate->setQualityValue(0.0);
  newEstimate->setQualityValue(0.0);
  estimate->setQualityValue(0.0);
  estimate->setObjectiveValue(9999999999.0);

  // Init estimates, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)
  if(programParameters->getOutputMode() > 1) cout << "\nCalculating startvalues, so that f(middleEstimate) < f(leftEstimate) and f(middleEstimate) < f(rightEstimate)" << endl;
  leftEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.0);
  if((leftEstimate->getD1() + leftEstimate->getD2() + leftEstimate->getD3() + leftEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationMEForPairs(programParameters,leftDistribution,programParameters->getNumberOfSimulations(),leftEstimate,leftEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 = " << (leftEstimate->getD1() + leftEstimate->getD2() + leftEstimate->getD3() + leftEstimate->getD4()) << " > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,leftEstimate,leftDistribution,targetHistogramPlot,lowerBound,upperBound);
  middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.5 * maximalSigmaE);
  if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationMEForPairs(programParameters,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 = " << (middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) << " > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
  rightEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,maximalSigmaE);
  if((rightEstimate->getD1() + rightEstimate->getD2() + rightEstimate->getD3() + rightEstimate->getD4()) < programParameters->getMaximalD()){
    threeTree->simulationMEForPairs(programParameters,rightDistribution,programParameters->getNumberOfSimulations(),rightEstimate,rightEstimate->getResults(),rnd,gammarnd);
  } else {
    cout << "d_1 + d_2 + d_3 + d_4 = " << (rightEstimate->getD1() + rightEstimate->getD2() + rightEstimate->getD3() + rightEstimate->getD4()) << " > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
    stop = true;
  }
  mEObjectiveFunction(programParameters,rightEstimate,rightDistribution,targetHistogramPlot,lowerBound,upperBound);
  if(programParameters->getOutputMode() > 1){
    cout << "\nLeft estimate:   " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
    cout << "Middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
    cout << "Right estimate:   " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
  }

  // Shift the interval to find a minimum
  while(!stop && !(leftEstimate->getObjectiveValue() > middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()) && ((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality()))){

    // left < middle < right
    if(leftEstimate->getObjectiveValue() <= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() <= rightEstimate->getObjectiveValue()) intervalShiftMode = 1;
    // left > middle > right
    if(leftEstimate->getObjectiveValue() >= middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() >= rightEstimate->getObjectiveValue()) intervalShiftMode = 2;
    // middle > (left && right)
    if(leftEstimate->getObjectiveValue() < middleEstimate->getObjectiveValue() && middleEstimate->getObjectiveValue() > rightEstimate->getObjectiveValue()) intervalShiftMode = 3;

    // left <= middle <= right
    if(intervalShiftMode == 1){
      cout << "\nLeft estimate <= Middle estimate <= Right estimate" << endl;
      copyEstimate(middleEstimate,rightEstimate);
      middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
    }
    // left >= middle >= right
    if(intervalShiftMode == 2){
      cout << "\nLeft estimate >= Middle estimate >= Right estimate" << endl;
      copyEstimate(middleEstimate,leftEstimate);
      middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
    }
    // middle > (left && right)
    if(intervalShiftMode == 3){
      cout << "\nMiddle estimate > (Left estimate && Right estimate)" << endl;
      if(leftEstimate->getObjectiveValue() < rightEstimate->getObjectiveValue()){
        // left < right
        copyEstimate(middleEstimate,rightEstimate);
        middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,0.5 * rightEstimate->getSigmaE());
      } else {
        // left > right
        copyEstimate(middleEstimate,leftEstimate);
        middleEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,(leftEstimate->getSigmaE() + rightEstimate->getSigmaE()) / 2.0);
      }
    }
    // Simulation for new middle estimate
    if((middleEstimate->getD1() + middleEstimate->getD2() + middleEstimate->getD3() + middleEstimate->getD4()) < programParameters->getMaximalD()){
      threeTree->simulationMEForPairs(programParameters,middleDistribution,programParameters->getNumberOfSimulations(),middleEstimate,middleEstimate->getResults(),rnd,gammarnd);
    } else {
      cout << "d_1 + d_2 + d_3 + d_4 > " << programParameters->getMaximalD() << " (Stop of the estimation process)" << endl;
      stop = true;
    }
    mEObjectiveFunction(programParameters,middleEstimate,middleDistribution,targetHistogramPlot,lowerBound,upperBound);
    if(programParameters->getOutputMode() > 1){
      cout << "\nNew left estimate:   " << leftEstimate->getSigmaE() << "\tDistance: " << leftEstimate->getObjectiveValue() << endl;
      cout << "New middle estimate:   " << middleEstimate->getSigmaE() << "\tDistance: " << middleEstimate->getObjectiveValue() << endl;
      cout << "New right estimate:   " << rightEstimate->getSigmaE() << "\tDistance: " << rightEstimate->getObjectiveValue() << endl;
    }
  }

  // Brents method for finding the minimum
  if(!stop){
    if(programParameters->getOutputMode() > 1) cout << "\nInitial values found. Start Brents Method." << endl;
    int i = 0;
    while((rightEstimate->getSigmaE() - leftEstimate->getSigmaE()) > (1.0 - programParameters->getQuality())){
      i++;
      if(rnd->getUniformDistributedNumber(0.0,1.0) < 0.5){
        newError = middleEstimate->getSigmaE() - 0.38197 * (middleEstimate->getSigmaE() - leftEstimate->getSigmaE());
        newEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationMEForPairs(programParameters,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,leftEstimate);
        } else {
          copyEstimate(middleEstimate,rightEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      } else {
        newError = middleEstimate->getSigmaE() + 0.38197 * (rightEstimate->getSigmaE() - middleEstimate->getSigmaE());
        newEstimate->setParametersMEExtremeGivenSigmaE(programParameters,targetValues,newError);
        threeTree->simulationMEForPairs(programParameters,newDistribution,programParameters->getNumberOfSimulations(),newEstimate,newEstimate->getResults(),rnd,gammarnd);
        mEObjectiveFunction(programParameters,newEstimate,newDistribution,targetHistogramPlot,lowerBound,upperBound);
        if(middleEstimate->getObjectiveValue() < newEstimate->getObjectiveValue()){
          copyEstimate(newEstimate,rightEstimate);
        } else {
          copyEstimate(middleEstimate,leftEstimate);
          copyEstimate(newEstimate,middleEstimate);
        }
      }
      if(programParameters->getOutputMode() > 1){
        cout << "\n----------------------------------------------------------------" << endl;
        cout << "\nIteration " << i;
        cout << "\nNew sigma_e estimate:  " << newEstimate->getSigmaE() << " \tDistance: " << newEstimate->getObjectiveValue() << endl;
        cout << "\nNew left estimate:   " << leftEstimate->getSigmaE() << " \tDistance: " << leftEstimate->getObjectiveValue() << endl;
        cout << "New middle estimate: " << middleEstimate->getSigmaE() << " \tDistance: " << middleEstimate->getObjectiveValue() << endl;
        cout << "New right estimate:  " << rightEstimate->getSigmaE() << " \tDistance: " << rightEstimate->getObjectiveValue() << endl;
      }
    }
    // Determining best estimate and set its quality equal 1.
    if(estimate->getObjectiveValue() > leftEstimate->getObjectiveValue()){
      copyEstimate(leftEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > middleEstimate->getObjectiveValue()){
      copyEstimate(middleEstimate,estimate);
    }
    if(estimate->getObjectiveValue() > rightEstimate->getObjectiveValue()){
      copyEstimate(rightEstimate,estimate);
    }
  }

  if(!stop){
    estimate->setQualityValue(1.0);
  } else {
    estimate->setQualityValue(0.0);
  }
}

// mEObjectiveFunction: Chi^2-function of the M&E-models
void Optimizer::mEObjectiveFunction(ProgramParameters *programParameters,ModelParameters *modelParameters,double modelParametersDistribution[],double targetHistogramPlot[],double lowerBound,double upperBound){
  GraphicTools gs = GraphicTools();
  double* modelParametersHistogramPlot = new double[programParameters->getNumberOfBins()];

  gs.calcHistogramPlot(modelParametersDistribution,programParameters->getNumberOfSimulations(),modelParametersHistogramPlot,programParameters->getNumberOfBins(),lowerBound,upperBound);

  long double distance = 0;
  for(long i = 0; i < programParameters->getNumberOfBins(); i++){
    if(modelParametersHistogramPlot[i] > 0){
      distance = distance + (pow(targetHistogramPlot[i] - modelParametersHistogramPlot[i],2.0) / modelParametersHistogramPlot[i]);
    } else {
      distance = distance + (pow((targetHistogramPlot[i] + 1.0) - (modelParametersHistogramPlot[i] + 1.0),2.0) / (modelParametersHistogramPlot[i] + 1.0));
    }
  }
  distance = pow(distance,0.5);
  modelParameters->setObjectiveValue(distance);
  modelParameters->setQualityValue(1.0 / (1.0 + (distance / (programParameters->getNumberOfSimulations()))));
}

// copyEstimate: Copies the values from the source estimate to the destination estimate
void Optimizer::copyEstimate(ModelParameters *source,ModelParameters *destination){
  destination->setAlpha(source->getAlpha());
  destination->setBetaOrSigmaM(source->getBetaOrSigmaM());
  destination->setSigmaE(source->getSigmaE());
  destination->setD1(source->getD1());
  destination->setD2(source->getD2());
  destination->setD3(source->getD3());
  destination->setD4(source->getD4());
  destination->setObjectiveValue(source->getObjectiveValue());
  destination->setQualityValue(source->getQualityValue());

  destination->getResults()->setVarianceWithin1(source->getResults()->getVarianceWithin1());
  destination->getResults()->setVarianceWithin2(source->getResults()->getVarianceWithin2());
  destination->getResults()->setVarianceWithin3(source->getResults()->getVarianceWithin3());
  destination->getResults()->setVariance12(source->getResults()->getVariance12());
  destination->getResults()->setVariance13(source->getResults()->getVariance13());
  destination->getResults()->setVariance23(source->getResults()->getVariance23());
  destination->getResults()->setSkewness12(source->getResults()->getSkewness12());
  destination->getResults()->setSkewness13(source->getResults()->getSkewness13());
  destination->getResults()->setSkewness23(source->getResults()->getSkewness23());
  destination->getResults()->setKurtosis12(source->getResults()->getKurtosis12());
  destination->getResults()->setKurtosis13(source->getResults()->getKurtosis13());
  destination->getResults()->setKurtosis23(source->getResults()->getKurtosis23());
  destination->getResults()->setS1iSkew(source->getResults()->getS1iSkew());
  destination->getResults()->setS2iSkew(source->getResults()->getS2iSkew());
  destination->getResults()->setS1iPearsonsSkew(source->getResults()->getS1iPearsonsSkew());
  destination->getResults()->setS2iPearsonsSkew(source->getResults()->getS2iPearsonsSkew());
  destination->getResults()->setOverallS1iCounter(source->getResults()->getOverallS1iCounter());
  destination->getResults()->setOverallS2iCounter(source->getResults()->getOverallS2iCounter());
  destination->getResults()->setSide(source->getResults()->getSide());
  destination->getResults()->setMirrored(source->getResults()->getMirrored());
}
