/***************************************************************************
                          optimizer.h  -  description
                             -------------------
    begin                : Thu Apr 28 2005
    copyright            : (C) 2005 by Michael Rosskopf
    email                : rosskopf@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*  Description:
An Optimizer-object has methods to estimate the parameters of an M-gamma model
and an M&E model (with Chi^2-method) from data sets.
*/

#ifndef OPTIMIZER_H
#define OPTIMIZER_H

#include "programparameters.h"
#include "modelparameters.h"
#include "distributionmeasures.h"
#include "data.h"
#include "threetree.h"
#include "randomnumbergenerator.h"
#include "random.h"
#include "graphictools.h"

/**
  *@author Michael Rosskopf
  */

class Optimizer {

  public: 

  // Constructors and destructors
  Optimizer(ProgramParameters*);
  ~Optimizer();

  // getEstimate: Returns the result of the previous parameter estimation
  ModelParameters* getEstimate();

  // setSideAndMirrored: Only used for the M-gamma model
  //                     All estimate buffer used by optimizer are setted to the used distribution
  //                     (normal gamma distribution ore mirrored one)
  void setSideAndMirrored(DistributionMeasures*);

  // estimateMGammaParameters: Estimates the parameters of the M-gamma model
  void estimateMGammaParameters(ProgramParameters*,Data*,ThreeTree*,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // mGammaObjectiveFunction: Objective function of the optimisation method of the M-gamma model
  void mGammaObjectiveFunction(ProgramParameters*,ModelParameters*,DistributionMeasures*);

  // estimateMENormalParameters: Estimates the parameters of the M&E-normal model
  void estimateMENormalParameters(ProgramParameters*,Data*,ThreeTree*,double[],double,double,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // estimateMENormalParametersForPairs: Estimates the parameters of the M&E-normal model for a pair of individuals
  void estimateMENormalParametersForPairs(ProgramParameters*,ThreeTree*,double[],double,double,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // estimateMEExtremeParameters: Estimates the parameters of the M&E-extreme model
  void estimateMEExtremeParameters(ProgramParameters*,Data*,ThreeTree*,double[],double,double,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // estimateMEExtremeParametersForPairs: Estimates the parameters of the M&E-extreme model for a pair of individuals
  void estimateMEExtremeParametersForPairs(ProgramParameters*,ThreeTree*,double[],double,double,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // mEObjectiveFunction: Chi^2-function of the M&E-models
  void mEObjectiveFunction(ProgramParameters*,ModelParameters*,double[],double[],double,double);

  // copyEstimate: Copy a model parameter estimate
  void copyEstimate(ModelParameters*,ModelParameters*);

  private:

  ModelParameters* leftEstimate;   // Buffer used by the Brents method
  ModelParameters* middleEstimate; // Buffer used by the Brents method
  ModelParameters* rightEstimate;  // Buffer used by the Brents method
  ModelParameters* newEstimate;    // Buffer used by the Brents method
  ModelParameters* estimate;       // Stores the result of a parameter estimation
  double* leftDistribution;        // Buffer used by the Brents method
  double* middleDistribution;      // Buffer used by the Brents method
  double* rightDistribution;       // Buffer used by the Brents method
  double* newDistribution;         // Buffer used by the Brents method
  double* distribution;            // Stores the gene expression distribution of the estimate
};

#endif
