/***************************************************************************
 *   Copyright (C) 2007 by Michael Rosskopf                                *
 *   rosskopf@cs.uni-duesseldorf.de                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*  Description:
A ProgramParameter-object reads the program parameters from the config.txt-file
which is used to specify the program execution.
*/

#ifndef PARAMETERS_H
#define PARAMETERS_H


#include <iostream.h>
#include <fstream.h>

/**
  *@author Michael Rosskopf
  */

class ProgramParameters{

  public: 

  // Constructors and destructors
  ProgramParameters(char*,char*);
  ~ProgramParameters();

  // get-methods
  int getMode();
  char* getModelParameterFilename();
  char* getDataTableFilename();
  int getSample1();
  int getSample2();
  int getSample3();
  int getAnalysisType();
  long getNumberOfGenes();
  long getNumberOfSimulations();
  long getNumberOfRuns();
  int getMED();
  double getQuality();
  int getOutputPrecision();
  int getOutputMode();
  double getStandardLowerBoundAlpha();
  double getMinimalAlpha();
  double getStandardUpperBoundAlpha();
  double getMaximalAlpha();
  int getSkewType();
  int getNumberOfSets();
  int getNonMutationalEffects();
  int getMethod();
  int getNumberOfBins();
  double getMaximalD();
  int getRepeats();
  double getSigmaMLo();
  double getSigmaMHi();
  double getSigmaELo();
  double getSigmaEHi();
  double getDLo();
  double getDHi();

  // set-methods
  void setMode(int);
  void setModelParameterFilename(char*);
  void setDataTableFilename(char*);
  void setSample1(int);
  void setSample2(int);
  void setSample3(int);
  void setAnalysisType(int);
  void setNumberOfGenes(long);
  void setNumberOfSimulations(long);
  void setNumberOfRuns(long);
  void setMED(int);
  void setQuality(double);
  void setOutputPrecision(int);
  void setOutputMode(int);
  void setStandardLowerBoundAlpha(double);
  void setMinimalAlpha(double);
  void setStandardUpperBoundAlpha(double);
  void setMaximalAlpha(double);
  void setSkewType(int);
  void setNumberOfSets(int);
  void setNonMutationalEffects(int);
  void setMethod(int);
  void setNumberOfBins(int);
  void setMaximalD(double);
  void setRepeats(int);
  void setSigmaMLo(double);
  void setSigmaMHi(double);
  void setSigmaELo(double);
  void setSigmaEHi(double);
  void setDLo(double);
  void setDHi(double);

  // to String:
  void toString();

  // checkProgramParameters: Returns true if program parameters are valid
  bool checkProgramParameters();

  private:

  int mode; // 1 -> Data generation
            // 2 -> Estimation on real data
            // 3 -> Estimation on bootstrap data
            // 4 -> Estimation on generated data

  char* modelParameterFilename; // Name of the used model parameter file
  char* dataTableFilename;      // Name of the used data set

  int sample1; // Label for individuals of sample 1
  int sample2; // Label for individuals of sample 2
  int sample3; // Label for individuals of sample 3

  int analysisType; //                                1 -> Mean of all pairwise comparisons
                    //                                       (between groups)
                    //                                2 -> Estimation for all pairs of samples
                    //                                       (within and between groups)
                    // Note: Estimation for all pairs can take a long time!

  long numberOfGenes;         // Number of genes if data is simulated
  long numberOfSimulations;   // Number of simulations of gene expression differences
  long numberOfRuns;          // Number of repetitions of the whole optimization process
  int mED;                    // 1 -> normal distribution
                              // 2 -> extremevalue distribution
                              // 3 -> gamma distribution

  // Optimization parameters
  double quality;             // Specifies quality (0-1) (used as a stop criterion)

  int outputPrecision;        // Number of digits in the output

  // Parameter for the screen output
  int outputMode; // 1 = No output, 2 = Output, 3 = Output and wait (fixed)

  // PROGRAM PARAMETERS RELATED TO THE M-GAMMA MODEL
  //
  double standardLowerBoundAlpha; // Lower bound for random initialisation of alpha of the optimisation method for the M-gamma model
  double minimalAlpha; // Minimal alpha of the search space of the optimisation method for the M-gamma model
  double standardUpperBoundAlpha; // Upper bound for random initialisation of alpha of the optimisation method for the M-gamma model
  double maximalAlpha; // Maximal alpha of the search space of the optimisation method for the M-gamma model
  int skewType;               // 1 = Coefficient of skewness
                              // 2 = Pearsons skew type 2 (fixed)
  int numberOfSets;           // 1 = Choose the greater set
                              // 2 = Mean of SLI and SRI results (fixed)

  // PROGRAM PARAMETERS RELATED TO THE M&E MODELS
  //
  int nonMutationalEffects;   // 1 = No error
                              // 2 = Normal distributed error
  int method; // Method to optimise the parameters of the M&E model
              // 1 -> Brents method which optimizes Chi^2-function
              // 2 -> Downhill simplex method which maximises the likelihood

  int numberOfBins;           // Number of bins for discretisation 
                              // (M&E model with Chi^2-method)
  double maximalD;            // Maximal value for sum of all edge lengths (M&E model with Chi^2-method)

  int repeats; // Number of repeats of the downhill simplex method 
               // if maximum likelihood is used

  double sigmaMLo; // Lower bound for random initialisation of sigma_m for the downhill simplex method
  double sigmaMHi; // Upper bound for random initialisation of sigma_m for the downhill simplex method
  double sigmaELo; // Lower bound for random initialisation of sigma_e for the downhill simplex method
  double sigmaEHi; // Upper bound for random initialisation of sigma_e for the downhill simplex method
  double dLo;      // Lower bound for random initialisation of d for the downhill simplex method
  double dHi;      // Upper bound for random initialisation of d for the downhill simplex method
};

#endif
