/***************************************************************************
                          threetree.h  -  description
                             -------------------
    begin                : Fri Dec 17 2004
    copyright            : (C) 2004 by Michael Rosskopf
    email                : rosskopf@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*  Description:
ThreeTree implements a hard-coded tree with 3 leaves used
for simulation of data corresponding to the model.

Please note that for some models only the edges d_1 and d_2 are simulated.

             root
              /\
             /  \
        d_4 /    \
           /      \
          /        \
         /\         \ 
        /  \         \
   d_1 /    \ d_2     \  d_3
      /      \         \
     /        \         \ 
 leftLeaf  rightLeaf  outLeaf

*/

#ifndef THREETREE_H
#define THREETREE_H

#include "programparameters.h"
#include "modelparameters.h"
#include "distributionmeasures.h"
#include "data.h"
#include "mathfunctions.h"
#include "randomnumbergenerator.h"
#include "random.h"

/**
  *@author Michael Rosskopf
  */

class ThreeTree {

  public: 

  // Constructors and destructors
  ThreeTree();
  ~ThreeTree();

  // simulationMGamma: Simulates data according to the M-gamma model
  // in order to estimate the s1i- and s2i-skewness
  void simulationMGamma(ProgramParameters*,Data*,long,ModelParameters*,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // simulationME: Simulates data according to the M&E model
  // in order to calculate the Chi^2-distance to analysed real data
  void simulationME(ProgramParameters*,Data*,double[],long,ModelParameters*,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // simulationMEForPairs: Simulates data according to the M&E model
  // in order to calculate the Chi^2-distance to pairs of individuals of analysed real data
  void simulationMEForPairs(ProgramParameters*,double[],long,ModelParameters*,DistributionMeasures*,RandomNumberGenerator*,Random*);

  // generation: Generates a synthetic data set according to the model parameters
  void generation(ProgramParameters*,ModelParameters*,Data*,RandomNumberGenerator*,Random*);

  private:

  // Used buffer variables for the simulation processes
  double alpha;
  double betaOrSigmaM;
  double sigmaE;
  double d1;
  double d2;
  double d3;
  double d4;

  double root;
  double ancestor;   
  double leftLeaf;
  double rightLeaf;
  double outLeaf;
  double difference;

  long loopCounter;
  long n;
  double numberOfMutations;
  long i,j,k;

  long l_geneCounter;
  long r_geneCounter;
  long double l_sum;
  long double l_squaredSum;
  long double l_cubedSum;
  long double sliVar;
  long double r_sum;
  long double r_squaredSum;
  long double r_cubedSum;
  long double sriVar;
};

#endif
